#!/usr/bin/env python
"""nf-core: Helper tools for use with nf-core Nextflow pipelines."""

import logging
import os
import sys
from pathlib import Path

import rich
import rich.console
import rich.logging
import rich.traceback
import rich_click as click
from trogon import tui

from nf_core import __version__
from nf_core.commands_modules import (
    modules_bump_versions,
    modules_create,
    modules_info,
    modules_install,
    modules_lint,
    modules_list_local,
    modules_list_remote,
    modules_patch,
    modules_remove,
    modules_test,
    modules_update,
)
from nf_core.commands_pipelines import (
    pipelines_bump_version,
    pipelines_create,
    pipelines_create_logo,
    pipelines_create_params_file,
    pipelines_download,
    pipelines_launch,
    pipelines_lint,
    pipelines_list,
    pipelines_rocrate,
    pipelines_schema_build,
    pipelines_schema_docs,
    pipelines_schema_lint,
    pipelines_schema_validate,
    pipelines_sync,
)
from nf_core.commands_subworkflows import (
    subworkflows_create,
    subworkflows_info,
    subworkflows_install,
    subworkflows_lint,
    subworkflows_list_local,
    subworkflows_list_remote,
    subworkflows_remove,
    subworkflows_test,
    subworkflows_update,
)
from nf_core.components.components_utils import NF_CORE_MODULES_REMOTE
from nf_core.pipelines.download import DownloadError
from nf_core.utils import check_if_outdated, nfcore_logo, rich_force_colors, setup_nfcore_dir

# Set up logging as the root logger
# Submodules should all traverse back to this
log = logging.getLogger()

# Set up .nfcore directory for storing files between sessions
setup_nfcore_dir()

# Set up nicer formatting of click cli help messages
click.rich_click.MAX_WIDTH = 100
click.rich_click.USE_RICH_MARKUP = True
click.rich_click.COMMAND_GROUPS = {
    "nf-core": [
        {
            "name": "Commands",
            "commands": [
                "pipelines",
                "modules",
                "subworkflows",
                "interface",
            ],
        },
    ],
    "nf-core pipelines": [
        {
            "name": "For users",
            "commands": ["list", "launch", "download", "create-params-file"],
        },
        {
            "name": "For developers",
            "commands": ["create", "lint", "bump-version", "sync", "schema", "rocrate", "create-logo"],
        },
    ],
    "nf-core modules": [
        {
            "name": "For pipelines",
            "commands": ["list", "info", "install", "update", "remove", "patch"],
        },
        {
            "name": "Developing new modules",
            "commands": ["create", "lint", "test", "bump-versions"],
        },
    ],
    "nf-core subworkflows": [
        {
            "name": "For pipelines",
            "commands": ["list", "info", "install", "update", "remove"],
        },
        {
            "name": "Developing new subworkflows",
            "commands": ["create", "lint", "test"],
        },
    ],
    "nf-core pipelines schema": [{"name": "Schema commands", "commands": ["validate", "build", "lint", "docs"]}],
}
click.rich_click.OPTION_GROUPS = {
    "nf-core modules list local": [{"options": ["--dir", "--json", "--help"]}],
}

# Set up rich stderr console
stderr = rich.console.Console(stderr=True, force_terminal=rich_force_colors())
stdout = rich.console.Console(force_terminal=rich_force_colors())

# Set up the rich traceback
rich.traceback.install(console=stderr, width=200, word_wrap=True, extra_lines=1)


# Define exceptions for which no traceback should be printed,
# because they are actually preliminary, but intended program terminations.
# (Custom exceptions are cleaner than `sys.exit(1)`, which we used before)
def selective_traceback_hook(exctype, value, traceback):
    if exctype in {DownloadError, UserWarning, ValueError}:  # extend set as needed
        log.error(value)
    else:
        # print the colored traceback for all other exceptions with rich as usual
        stderr.print(rich.traceback.Traceback.from_exception(exctype, value, traceback))


sys.excepthook = selective_traceback_hook


# Define callback function to normalize the case of click arguments,
# which is used to make the module/subworkflow names, provided by the
# user on the cli, case insensitive.
def normalize_case(ctx, param, component_name):
    if component_name is not None:
        return component_name.casefold()


# Define a custom click group class to sort options and commands in the help message
# TODO: Remove this class and use COMMANDS_BEFORE_OPTIONS when rich-click is updated
# See https://github.com/ewels/rich-click/issues/200 for more information
class CustomRichGroup(click.RichGroup):
    def format_options(self, ctx, formatter) -> None:
        from rich_click.rich_help_rendering import get_rich_options

        self.format_commands(ctx, formatter)
        get_rich_options(self, ctx, formatter)


def run_nf_core():
    # print nf-core header if environment variable is not set
    if os.environ.get("_NF_CORE_COMPLETE") is None:
        # Print nf-core header
        stderr.print("\n")
        for line in nfcore_logo:
            stderr.print(line, highlight=False)
        stderr.print(
            f"\n[grey39]    nf-core/tools version {__version__} - [link=https://nf-co.re]https://nf-co.re[/]",
            highlight=False,
        )
        try:
            is_outdated, _, remote_vers = check_if_outdated()
            if is_outdated:
                stderr.print(
                    f"[bold bright_yellow]    There is a new version of nf-core/tools available! ({remote_vers})",
                    highlight=False,
                )
        except Exception as e:
            log.debug(f"Could not check latest version: {e}")
        stderr.print("\n")
    # Launch the click cli
    nf_core_cli(auto_envvar_prefix="NFCORE")


@tui(
    command="interface",
    help="Launch the nf-core interface",
)
@click.group(context_settings=dict(help_option_names=["-h", "--help"]), cls=CustomRichGroup)
@click.version_option(__version__)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Print verbose output to the console.",
)
@click.option("--hide-progress", is_flag=True, default=False, help="Don't show progress bars.")
@click.option("-l", "--log-file", help="Save a verbose log to a file.", metavar="<filename>")
@click.pass_context
def nf_core_cli(ctx, verbose, hide_progress, log_file):
    """
    nf-core/tools provides a set of helper tools for use with nf-core Nextflow pipelines.

    It is designed for both end-users running pipelines and also developers creating new pipelines.
    """
    # Set the base logger to output DEBUG
    log.setLevel(logging.DEBUG)

    # Set up logs to the console
    log.addHandler(
        rich.logging.RichHandler(
            level=logging.DEBUG if verbose else logging.INFO,
            console=rich.console.Console(stderr=True, force_terminal=rich_force_colors()),
            show_time=False,
            show_path=verbose,  # True if verbose, false otherwise
            markup=True,
        )
    )

    # don't show rich debug logging in verbose mode
    rich_logger = logging.getLogger("rich")
    rich_logger.setLevel(logging.INFO)

    # Set up logs to a file if we asked for one
    if log_file:
        log_fh = logging.FileHandler(log_file, encoding="utf-8")
        log_fh.setLevel(logging.DEBUG)
        log_fh.setFormatter(logging.Formatter("[%(asctime)s] %(name)-20s [%(levelname)-7s]  %(message)s"))
        log.addHandler(log_fh)

    ctx.obj = {
        "verbose": verbose,
        "hide_progress": hide_progress or verbose,  # Always hide progress bar with verbose logging
    }


# nf-core pipelines subcommands
@nf_core_cli.group()
@click.pass_context
def pipelines(ctx):
    """
    Commands to manage nf-core pipelines.
    """
    # ensure that ctx.obj exists and is a dict (in case `cli()` is called
    # by means other than the `if` block below)
    ctx.ensure_object(dict)


# nf-core pipelines create
@pipelines.command("create")
@click.pass_context
@click.option(
    "-n",
    "--name",
    type=str,
    help="The name of your new pipeline",
)
@click.option("-d", "--description", type=str, help="A short description of your pipeline")
@click.option("-a", "--author", type=str, help="Name of the main author(s)")
@click.option("--version", type=str, default="1.0.0dev", help="The initial version number to use")
@click.option("-f", "--force", is_flag=True, default=False, help="Overwrite output directory if it already exists")
@click.option("-o", "--outdir", help="Output directory for new pipeline (default: pipeline name)")
@click.option("-t", "--template-yaml", help="Pass a YAML file to customize the template")
@click.option(
    "--organisation",
    type=str,
    default="nf-core",
    help="The name of the GitHub organisation where the pipeline will be hosted (default: nf-core)",
)
def command_pipelines_create(ctx, name, description, author, version, force, outdir, template_yaml, organisation):
    """
    Create a new pipeline using the nf-core template.
    """
    pipelines_create(ctx, name, description, author, version, force, outdir, template_yaml, organisation)


# nf-core pipelines lint
@pipelines.command("lint")
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    help=r"Pipeline directory [dim]\[default: current working directory][/]",
)
@click.option(
    "--release",
    is_flag=True,
    default=Path(os.environ.get("GITHUB_REF", "").strip(" '\"")).parent.name in ["master", "main"]
    and os.environ.get("GITHUB_REPOSITORY", "").startswith("nf-core/")
    and not os.environ.get("GITHUB_REPOSITORY", "") == "nf-core/tools",
    help="Execute additional checks for release-ready workflows.",
)
@click.option(
    "-f",
    "--fix",
    type=str,
    metavar="<test>",
    multiple=True,
    help="Attempt to automatically fix specified lint test",
)
@click.option(
    "-k",
    "--key",
    type=str,
    metavar="<test>",
    multiple=True,
    help="Run only these lint tests",
)
@click.option("-p", "--show-passed", is_flag=True, help="Show passing tests on the command line")
@click.option("-i", "--fail-ignored", is_flag=True, help="Convert ignored tests to failures")
@click.option("-w", "--fail-warned", is_flag=True, help="Convert warn tests to failures")
@click.option(
    "--markdown",
    type=str,
    metavar="<filename>",
    help="File to write linting results to (Markdown)",
)
@click.option(
    "--json",
    type=str,
    metavar="<filename>",
    help="File to write linting results to (JSON)",
)
@click.option(
    "--sort-by",
    type=click.Choice(["module", "test"]),
    default="test",
    help="Sort lint output by module or test name.",
    show_default=True,
)
@click.pass_context
def command_pipelines_lint(
    ctx,
    directory,
    release,
    fix,
    key,
    show_passed,
    fail_ignored,
    fail_warned,
    markdown,
    json,
    sort_by,
):
    """
    Check pipeline code against nf-core guidelines.
    """
    pipelines_lint(ctx, directory, release, fix, key, show_passed, fail_ignored, fail_warned, markdown, json, sort_by)


# nf-core pipelines download
@pipelines.command("download")
@click.argument("pipeline", required=False, metavar="<pipeline name>")
@click.option(
    "-r",
    "--revision",
    multiple=True,
    help="Pipeline release to download. Multiple invocations are possible, e.g. `-r 1.1 -r 1.2`",
)
@click.option("-o", "--outdir", type=str, help="Output directory")
@click.option(
    "-x",
    "--compress",
    type=click.Choice(["tar.gz", "tar.bz2", "zip", "none"]),
    help="Archive compression type",
)
@click.option("-f", "--force", is_flag=True, default=False, help="Overwrite existing files")
@click.option(
    "-p",
    "--platform",
    is_flag=True,
    default=False,
    help="Download for Seqera Platform (formerly Nextflow Tower)",
)
@click.option(
    "-c",
    "--download-configuration",
    type=click.Choice(["yes", "no"]),
    default="no",
    help="Include configuration profiles in download. Not available with `--platform`",
)
@click.option(
    "--tag",
    multiple=True,
    help="Add custom alias tags to `--platform` downloads. For example, `--tag \"3.10=validated\"` adds the custom 'validated' tag to the 3.10 release.",
)
# -c changed to -s for consistency with other --container arguments, where it is always the first letter of the last word.
# Also -c might be used instead of -d for config in a later release, but reusing params for different options in two subsequent releases might be too error-prone.
@click.option(
    "-s",
    "--container-system",
    type=click.Choice(["none", "singularity"]),
    help="Download container images of required software.",
)
@click.option(
    "-l",
    "--container-library",
    multiple=True,
    help="Container registry/library or mirror to pull images from.",
)
@click.option(
    "-u",
    "--container-cache-utilisation",
    type=click.Choice(["amend", "copy", "remote"]),
    help="Utilise a `singularity.cacheDir` in the download process, if applicable.",
)
@click.option(
    "-i",
    "--container-cache-index",
    type=str,
    help="List of images already available in a remote `singularity.cacheDir`.",
)
@click.option(
    "-d",
    "--parallel-downloads",
    type=int,
    default=4,
    help="Number of parallel image downloads",
)
@click.pass_context
def command_pipelines_download(
    ctx,
    pipeline,
    revision,
    outdir,
    compress,
    force,
    platform,
    download_configuration,
    tag,
    container_system,
    container_library,
    container_cache_utilisation,
    container_cache_index,
    parallel_downloads,
):
    """
    Download a pipeline, nf-core/configs and pipeline singularity images.
    """
    pipelines_download(
        ctx,
        pipeline,
        revision,
        outdir,
        compress,
        force,
        platform,
        download_configuration,
        tag,
        container_system,
        container_library,
        container_cache_utilisation,
        container_cache_index,
        parallel_downloads,
    )


# nf-core pipelines create-params-file
@pipelines.command("create-params-file")
@click.argument("pipeline", required=False, metavar="<pipeline name>")
@click.option("-r", "--revision", help="Release/branch/SHA of the pipeline (if remote)")
@click.option(
    "-o",
    "--output",
    type=str,
    default="nf-params.yml",
    metavar="<filename>",
    help="Output filename. Defaults to `nf-params.yml`.",
)
@click.option("-f", "--force", is_flag=True, default=False, help="Overwrite existing files")
@click.option(
    "-x",
    "--show-hidden",
    is_flag=True,
    default=False,
    help="Show hidden params which don't normally need changing",
)
@click.pass_context
def command_pipelines_create_params_file(ctx, pipeline, revision, output, force, show_hidden):
    """
    Build a parameter file for a pipeline.
    """
    pipelines_create_params_file(ctx, pipeline, revision, output, force, show_hidden)


# nf-core pipelines launch
@pipelines.command("launch")
@click.argument("pipeline", required=False, metavar="<pipeline name>")
@click.option("-r", "--revision", help="Release/branch/SHA of the project to run (if remote)")
@click.option("-i", "--id", help="ID for web-gui launch parameter set")
@click.option(
    "-c",
    "--command-only",
    is_flag=True,
    default=False,
    help="Create Nextflow command with params (no params file)",
)
@click.option(
    "-o",
    "--params-out",
    type=click.Path(),
    default=os.path.join(os.getcwd(), "nf-params.json"),
    help="Path to save run parameters file",
)
@click.option(
    "-p",
    "--params-in",
    type=click.Path(exists=True),
    help="Set of input run params to use from a previous run",
)
@click.option(
    "-a",
    "--save-all",
    is_flag=True,
    default=False,
    help="Save all parameters, even if unchanged from default",
)
@click.option(
    "-x",
    "--show-hidden",
    is_flag=True,
    default=False,
    help="Show hidden params which don't normally need changing",
)
@click.option(
    "-u",
    "--url",
    type=str,
    default="https://nf-co.re/launch",
    help="Customise the builder URL (for development work)",
)
@click.pass_context
def command_pipelines_launch(
    ctx,
    pipeline,
    id,
    revision,
    command_only,
    params_in,
    params_out,
    save_all,
    show_hidden,
    url,
):
    """
    Launch a pipeline using a web GUI or command line prompts.
    """
    pipelines_launch(ctx, pipeline, id, revision, command_only, params_in, params_out, save_all, show_hidden, url)


# nf-core pipelines list
@pipelines.command("list")
@click.argument("keywords", required=False, nargs=-1, metavar="<filter keywords>")
@click.option(
    "-s",
    "--sort",
    type=click.Choice(["release", "pulled", "name", "stars"]),
    default="release",
    help="How to sort listed pipelines",
)
@click.option("--json", is_flag=True, default=False, help="Print full output as JSON")
@click.option("--show-archived", is_flag=True, default=False, help="Print archived workflows")
@click.pass_context
def command_pipelines_list(ctx, keywords, sort, json, show_archived):
    """
    List available nf-core pipelines with local info.
    """
    pipelines_list(ctx, keywords, sort, json, show_archived)


# nf-core pipelines rocrate
@pipelines.command("rocrate")
@click.argument(
    "pipeline_dir",
    type=click.Path(exists=True),
    default=Path.cwd(),
    required=True,
    metavar="<pipeline directory>",
)
@click.option(
    "-j",
    "--json_path",
    default=Path.cwd(),
    type=str,
    help="Path to save RO Crate metadata json file to",
)
@click.option("-z", "--zip_path", type=str, help="Path to save RO Crate zip file to")
@click.option(
    "-pv",
    "--pipeline_version",
    type=str,
    help="Version of pipeline to use for RO Crate",
    default="",
)
@click.pass_context
def rocrate(
    ctx,
    pipeline_dir: str,
    json_path: str,
    zip_path: str,
    pipeline_version: str,
):
    """
    Make an Research Object Crate
    """
    pipelines_rocrate(ctx, pipeline_dir, json_path, zip_path, pipeline_version)


# nf-core pipelines sync
@pipelines.command("sync")
@click.pass_context
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    help=r"Pipeline directory. [dim]\[default: current working directory][/]",
)
@click.option(
    "-b",
    "--from-branch",
    type=str,
    help="The git branch to use to fetch workflow variables.",
)
@click.option(
    "-p",
    "--pull-request",
    is_flag=True,
    default=False,
    help="Make a GitHub pull-request with the changes.",
)
@click.option(
    "--force_pr",
    is_flag=True,
    default=False,
    help="Force the creation of a pull-request, even if there are no changes.",
)
@click.option("-g", "--github-repository", type=str, help="GitHub PR: target repository.")
@click.option("-u", "--username", type=str, help="GitHub PR: auth username.")
@click.option("-t", "--template-yaml", help="Pass a YAML file to customize the template")
def command_pipelines_sync(
    ctx, directory, from_branch, pull_request, github_repository, username, template_yaml, force_pr
):
    """
    Sync a pipeline [cyan i]TEMPLATE[/] branch with the nf-core template.
    """
    pipelines_sync(ctx, directory, from_branch, pull_request, github_repository, username, template_yaml, force_pr)


# nf-core pipelines bump-version
@pipelines.command("bump-version")
@click.pass_context
@click.argument("new_version", required=True, metavar="<new version>")
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    help=r"Pipeline directory. [dim]\[default: current working directory][/]",
)
@click.option(
    "-n",
    "--nextflow",
    is_flag=True,
    default=False,
    help="Bump required nextflow version instead of pipeline version",
)
def command_pipelines_bump_version(ctx, new_version, directory, nextflow):
    """
    Update nf-core pipeline version number with `nf-core pipelines bump-version`.
    """
    pipelines_bump_version(ctx, new_version, directory, nextflow)


# nf-core pipelines create-logo
@pipelines.command("create-logo")
@click.argument("logo-text", metavar="<logo_text>")
@click.option("-d", "--dir", "directory", type=click.Path(), default=".", help="Directory to save the logo in.")
@click.option(
    "-n",
    "--name",
    type=str,
    help="Name of the output file (with or without '.png' suffix).",
)
@click.option(
    "--theme",
    type=click.Choice(["light", "dark"]),
    default="light",
    help="Theme for the logo.",
    show_default=True,
)
@click.option(
    "--width",
    type=int,
    default=2300,
    help="Width of the logo in pixels.",
    show_default=True,
)
@click.option(
    "--format",
    type=click.Choice(["png", "svg"]),
    default="png",
    help="Image format of the logo, either PNG or SVG.",
    show_default=True,
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Overwrite any files if they already exist",
)
def command_pipelines_create_logo(logo_text, directory, name, theme, width, format, force):
    """
    Generate a logo with the nf-core logo template.
    """
    pipelines_create_logo(logo_text, directory, name, theme, width, format, force)


# nf-core pipelines schema subcommands
@pipelines.group("schema")
def pipeline_schema():
    """
    Suite of tools for developers to manage pipeline schema.

    All nf-core pipelines should have a nextflow_schema.json file in their
    root directory that describes the different pipeline parameters.
    """
    pass


# nf-core pipelines schema validate
@pipeline_schema.command("validate")
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    help=r"Pipeline directory. [dim]\[default: current working directory][/]",
)
@click.argument("pipeline", required=True, metavar="<pipeline name>")
@click.argument("params", type=click.Path(exists=True), required=True, metavar="<JSON params file>")
def command_pipelines_schema_validate(directory, pipeline, params):
    """
    Validate a set of parameters against a pipeline schema.
    """
    if Path(directory, pipeline).exists():
        # this is a local pipeline
        pipeline = Path(directory, pipeline)

    pipelines_schema_validate(pipeline, params)


# nf-core pipelines schema build
@pipeline_schema.command("build")
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    help=r"Pipeline directory. [dim]\[default: current working directory][/]",
)
@click.option(
    "--no-prompts",
    is_flag=True,
    help="Do not confirm changes, just update parameters and exit",
)
@click.option(
    "--web-only",
    is_flag=True,
    help="Skip building using Nextflow config, just launch the web tool",
)
@click.option(
    "--url",
    type=str,
    default="https://nf-co.re/pipeline_schema_builder",
    help="Customise the builder URL (for development work)",
)
def command_pipelines_schema_build(directory, no_prompts, web_only, url):
    """
    Interactively build a pipeline schema from Nextflow params.
    """
    pipelines_schema_build(directory, no_prompts, web_only, url)


# nf-core pipelines schema lint
@pipeline_schema.command("lint")
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    help=r"Pipeline directory. [dim]\[default: current working directory][/]",
)
@click.argument(
    "schema_file",
    type=click.Path(exists=True),
    default="nextflow_schema.json",
    metavar="<pipeline schema>",
)
def command_pipelines_schema_lint(directory, schema_file):
    """
    Check that a given pipeline schema is valid.
    """
    pipelines_schema_lint(Path(directory, schema_file))


# nf-core pipelines schema docs
@pipeline_schema.command("docs")
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    help=r"Pipeline directory. [dim]\[default: current working directory][/]",
)
@click.argument(
    "schema_file",
    type=click.Path(exists=True),
    default="nextflow_schema.json",
    required=False,
    metavar="<pipeline schema>",
)
@click.option(
    "-o",
    "--output",
    type=str,
    metavar="<filename>",
    help="Output filename. Defaults to standard out.",
)
@click.option(
    "-x",
    "--format",
    type=click.Choice(["markdown", "html"]),
    default="markdown",
    help="Format to output docs in.",
)
@click.option("-f", "--force", is_flag=True, default=False, help="Overwrite existing files")
@click.option(
    "-c",
    "--columns",
    type=str,
    metavar="<columns_list>",
    help="CSV list of columns to include in the parameter tables (parameter,description,type,default,required,hidden)",
    default="parameter,description,type,default,required,hidden",
)
def command_pipelines_schema_docs(directory, schema_file, output, format, force, columns):
    """
    Outputs parameter documentation for a pipeline schema.
    """
    pipelines_schema_docs(Path(directory, schema_file), output, format, force, columns)


# nf-core modules subcommands
@nf_core_cli.group()
@click.option(
    "-g",
    "--git-remote",
    type=str,
    default=NF_CORE_MODULES_REMOTE,
    help="Remote git repo to fetch files from",
)
@click.option(
    "-b",
    "--branch",
    type=str,
    default=None,
    help="Branch of git repository hosting modules.",
)
@click.option(
    "-N",
    "--no-pull",
    is_flag=True,
    default=False,
    help="Do not pull in latest changes to local clone of modules repository.",
)
@click.pass_context
def modules(ctx, git_remote, branch, no_pull):
    """
    Commands to manage Nextflow DSL2 modules (tool wrappers).
    """
    # ensure that ctx.obj exists and is a dict (in case `cli()` is called
    # by means other than the `if` block below)
    ctx.ensure_object(dict)

    # Place the arguments in a context object
    ctx.obj["modules_repo_url"] = git_remote
    ctx.obj["modules_repo_branch"] = branch
    ctx.obj["modules_repo_no_pull"] = no_pull


# nf-core modules list subcommands
@modules.group("list")
@click.pass_context
def modules_list(ctx):
    """
    List modules in a local pipeline or remote repository.
    """
    pass


# nf-core modules list remote
@modules_list.command("remote")
@click.pass_context
@click.argument("keywords", required=False, nargs=-1, metavar="<filter keywords>")
@click.option("-j", "--json", is_flag=True, help="Print as JSON to stdout")
def command_modules_list_remote(ctx, keywords, json):
    """
    List modules in a remote GitHub repo [dim i](e.g [link=https://github.com/nf-core/modules]nf-core/modules[/])[/].
    """
    modules_list_remote(ctx, keywords, json)


# nf-core modules list local
@modules_list.command("local")
@click.pass_context
@click.argument("keywords", required=False, nargs=-1, metavar="<filter keywords>")
@click.option("-j", "--json", is_flag=True, help="Print as JSON to stdout")
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    help=r"Pipeline directory. [dim]\[default: Current working directory][/]",
)
def command_modules_list_local(ctx, keywords, json, directory):  # pylint: disable=redefined-builtin
    """
    List modules installed locally in a pipeline
    """
    modules_list_local(ctx, keywords, json, directory)


# nf-core modules install
@modules.command("install")
@click.pass_context
@click.argument("tool", type=str, callback=normalize_case, required=False, metavar="<tool> or <tool/subtool>")
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    help=r"Pipeline directory. [dim]\[default: current working directory][/]",
)
@click.option(
    "-p",
    "--prompt",
    is_flag=True,
    default=False,
    help="Prompt for the version of the module",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force reinstallation of module if it already exists",
)
@click.option("-s", "--sha", type=str, metavar="<commit sha>", help="Install module at commit SHA")
def command_modules_install(ctx, tool, directory, prompt, force, sha):
    """
    Install DSL2 modules within a pipeline.
    """
    modules_install(ctx, tool, directory, prompt, force, sha)


# nf-core modules update
@modules.command("update")
@click.pass_context
@click.argument("tool", type=str, callback=normalize_case, required=False, metavar="<tool> or <tool/subtool>")
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    help=r"Pipeline directory. [dim]\[default: current working directory][/]",
)
@click.option("-f", "--force", is_flag=True, default=False, help="Force update of module")
@click.option(
    "-p",
    "--prompt",
    is_flag=True,
    default=False,
    help="Prompt for the version of the module",
)
@click.option(
    "--limit-output",
    "limit_output",
    is_flag=True,
    default=False,
    help="Limit output to only the difference in main.nf",
)
@click.option("-s", "--sha", type=str, metavar="<commit sha>", help="Install module at commit SHA")
@click.option(
    "-a",
    "--all",
    "install_all",
    is_flag=True,
    default=False,
    help="Update all modules installed in pipeline",
)
@click.option(
    "-x/-y",
    "--preview/--no-preview",
    is_flag=True,
    default=None,
    help="Preview / no preview of changes before applying",
)
@click.option(
    "-D",
    "--save-diff",
    type=str,
    metavar="<filename>",
    default=None,
    help="Save diffs to a file instead of updating in place",
)
@click.option(
    "-u",
    "--update-deps",
    is_flag=True,
    default=False,
    help="Automatically update all linked modules and subworkflows without asking for confirmation",
)
def command_modules_update(
    ctx,
    tool,
    directory,
    force,
    prompt,
    sha,
    install_all,
    preview,
    save_diff,
    update_deps,
    limit_output,
):
    """
    Update DSL2 modules within a pipeline.
    """
    modules_update(ctx, tool, directory, force, prompt, sha, install_all, preview, save_diff, update_deps, limit_output)


# nf-core modules patch
@modules.command("patch")
@click.pass_context
@click.argument("tool", type=str, callback=normalize_case, required=False, metavar="<tool> or <tool/subtool>")
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    help=r"Pipeline directory. [dim]\[default: current working directory][/]",
)
@click.option("-r", "--remove", is_flag=True, default=False, help="Remove an existent patch file and regenerate it.")
def command_modules_patch(ctx, tool, directory, remove):
    """
    Create a patch file for minor changes in a module
    """
    modules_patch(ctx, tool, directory, remove)


# nf-core modules remove
@modules.command("remove")
@click.pass_context
@click.argument("tool", type=str, callback=normalize_case, required=False, metavar="<tool> or <tool/subtool>")
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    help=r"Pipeline directory. [dim]\[default: current working directory][/]",
)
def command_modules_remove(ctx, directory, tool):
    """
    Remove a module from a pipeline.
    """
    modules_remove(ctx, directory, tool)


# nf-core modules create
@modules.command("create")
@click.pass_context
@click.argument("tool", type=str, required=False, metavar="<tool> or <tool/subtool>")
@click.option("-d", "--dir", "directory", type=click.Path(exists=True), default=".", metavar="<directory>")
@click.option(
    "-a",
    "--author",
    type=str,
    metavar="<author>",
    help="Module author's GitHub username prefixed with '@'",
)
@click.option(
    "-l",
    "--label",
    type=str,
    metavar="<process label>",
    help="Standard resource label for process",
)
@click.option(
    "-m",
    "--meta",
    is_flag=True,
    default=False,
    help="Use Groovy meta map for sample information",
)
@click.option(
    "-n",
    "--no-meta",
    is_flag=True,
    default=False,
    help="Don't use meta map for sample information",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Overwrite any files if they already exist",
)
@click.option(
    "-c",
    "--conda-name",
    type=str,
    default=None,
    help="Name of the conda package to use",
)
@click.option(
    "-p",
    "--conda-package-version",
    type=str,
    default=None,
    help="Version of conda package to use",
)
@click.option(
    "-i",
    "--empty-template",
    is_flag=True,
    default=False,
    help="Create a module from the template without TODOs or examples",
)
@click.option(
    "--migrate-pytest",
    is_flag=True,
    default=False,
    help="Migrate a module with pytest tests to nf-test",
)
def command_modules_create(
    ctx,
    tool,
    directory,
    author,
    label,
    meta,
    no_meta,
    force,
    conda_name,
    conda_package_version,
    empty_template,
    migrate_pytest,
):
    """
    Create a new DSL2 module from the nf-core template.
    """
    modules_create(
        ctx,
        tool,
        directory,
        author,
        label,
        meta,
        no_meta,
        force,
        conda_name,
        conda_package_version,
        empty_template,
        migrate_pytest,
    )


# nf-core modules test
@modules.command("test")
@click.pass_context
@click.argument("tool", type=str, callback=normalize_case, required=False, metavar="<tool> or <tool/subtool>")
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Print verbose output to the console. Sets `--debug` inside the nf-test command.",
)
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    metavar="<nf-core/modules directory>",
)
@click.option(
    "-p",
    "--no-prompts",
    is_flag=True,
    default=False,
    help="Use defaults without prompting",
)
@click.option("-u", "--update", is_flag=True, default=False, help="Update existing snapshots")
@click.option(
    "-o",
    "--once",
    is_flag=True,
    default=False,
    help="Run tests only once. Don't check snapshot stability",
)
@click.option(
    "--profile",
    type=click.Choice(["docker", "singularity", "conda"]),
    default=None,
    help="Run tests with a specific profile",
)
@click.option(
    "--migrate-pytest",
    is_flag=True,
    default=False,
    help="Migrate a module with pytest tests to nf-test",
)
def command_modules_test(ctx, tool, directory, no_prompts, update, once, profile, migrate_pytest, verbose):
    """
    Run nf-test for a module.
    """
    if verbose:
        ctx.obj["verbose"] = verbose
    modules_test(ctx, tool, directory, no_prompts, update, once, profile, migrate_pytest)


# nf-core modules lint
@modules.command("lint")
@click.pass_context
@click.argument("tool", type=str, callback=normalize_case, required=False, metavar="<tool> or <tool/subtool>")
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    metavar="<pipeline/modules directory>",
)
@click.option(
    "-r",
    "--registry",
    type=str,
    metavar="<registry>",
    default=None,
    help="Registry to use for containers. If not specified it will use docker.registry value in the nextflow.config file",
)
@click.option(
    "-k",
    "--key",
    type=str,
    metavar="<test>",
    multiple=True,
    help="Run only these lint tests",
)
@click.option("-a", "--all", is_flag=True, help="Run on all modules")
@click.option("-w", "--fail-warned", is_flag=True, help="Convert warn tests to failures")
@click.option("--local", is_flag=True, help="Run additional lint tests for local modules")
@click.option("--passed", is_flag=True, help="Show passed tests")
@click.option(
    "--sort-by",
    type=click.Choice(["module", "test"]),
    default="test",
    help="Sort lint output by module or test name.",
    show_default=True,
)
@click.option(
    "--fix-version",
    is_flag=True,
    help="Fix the module version if a newer version is available",
)
@click.option("--fix", is_flag=True, help="Fix all linting tests if possible.")
def command_modules_lint(
    ctx, tool, directory, registry, key, all, fail_warned, local, passed, sort_by, fix_version, fix
):
    """
    Lint one or more modules in a directory.
    """
    modules_lint(ctx, tool, directory, registry, key, all, fail_warned, local, passed, sort_by, fix_version, fix)


# nf-core modules info
@modules.command("info")
@click.pass_context
@click.argument("tool", type=str, callback=normalize_case, required=False, metavar="<tool> or <tool/subtool>")
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    help=r"Pipeline directory. [dim]\[default: Current working directory][/]",
)
def command_modules_info(ctx, tool, directory):
    """
    Show developer usage information about a given module.
    """
    modules_info(ctx, tool, directory)


# nf-core modules bump-versions
@modules.command("bump-versions")
@click.pass_context
@click.argument("tool", type=str, callback=normalize_case, required=False, metavar="<tool> or <tool/subtool>")
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    metavar="<nf-core/modules directory>",
)
@click.option("-a", "--all", is_flag=True, help="Run on all modules")
@click.option("-s", "--show-all", is_flag=True, help="Show up-to-date modules in results too")
def command_modules_bump_versions(ctx, tool, directory, all, show_all):
    """
    Bump versions for one or more modules in a clone of
    the nf-core/modules repo.
    """
    modules_bump_versions(ctx, tool, directory, all, show_all)


# nf-core subworkflows click command
@nf_core_cli.group()
@click.option(
    "-g",
    "--git-remote",
    type=str,
    default=NF_CORE_MODULES_REMOTE,
    help="Remote git repo to fetch files from",
)
@click.option(
    "-b",
    "--branch",
    type=str,
    default=None,
    help="Branch of git repository hosting modules.",
)
@click.option(
    "-N",
    "--no-pull",
    is_flag=True,
    default=False,
    help="Do not pull in latest changes to local clone of modules repository.",
)
@click.pass_context
def subworkflows(ctx, git_remote, branch, no_pull):
    """
    Commands to manage Nextflow DSL2 subworkflows (tool wrappers).
    """
    # ensure that ctx.obj exists and is a dict (in case `cli()` is called
    # by means other than the `if` block below)
    ctx.ensure_object(dict)

    # Place the arguments in a context object
    ctx.obj["modules_repo_url"] = git_remote
    ctx.obj["modules_repo_branch"] = branch
    ctx.obj["modules_repo_no_pull"] = no_pull


# nf-core subworkflows create
@subworkflows.command("create")
@click.pass_context
@click.argument("subworkflow", type=str, required=False, metavar="subworkflow name")
@click.option("-d", "--dir", "directory", type=click.Path(exists=True), default=".", metavar="<directory>")
@click.option(
    "-a",
    "--author",
    type=str,
    metavar="<author>",
    help="Module author's GitHub username prefixed with '@'",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Overwrite any files if they already exist",
)
@click.option(
    "--migrate-pytest",
    is_flag=True,
    default=False,
    help="Migrate a module with pytest tests to nf-test",
)
def command_subworkflows_create(ctx, subworkflow, directory, author, force, migrate_pytest):
    """
    Create a new subworkflow from the nf-core template.
    """
    subworkflows_create(ctx, subworkflow, directory, author, force, migrate_pytest)


# nf-core subworkflows test
@subworkflows.command("test")
@click.pass_context
@click.argument("subworkflow", type=str, callback=normalize_case, required=False, metavar="subworkflow name")
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    metavar="<nf-core/modules directory>",
)
@click.option(
    "-p",
    "--no-prompts",
    is_flag=True,
    default=False,
    help="Use defaults without prompting",
)
@click.option("-u", "--update", is_flag=True, default=False, help="Update existing snapshots")
@click.option(
    "-o",
    "--once",
    is_flag=True,
    default=False,
    help="Run tests only once. Don't check snapshot stability",
)
@click.option(
    "--profile",
    type=click.Choice(["docker", "singularity", "conda"]),
    default=None,
    help="Run tests with a specific profile",
)
@click.option(
    "--migrate-pytest",
    is_flag=True,
    default=False,
    help="Migrate a subworkflow with pytest tests to nf-test",
)
def command_subworkflows_test(ctx, subworkflow, directory, no_prompts, update, once, profile, migrate_pytest):
    """
    Run nf-test for a subworkflow.
    """
    subworkflows_test(ctx, subworkflow, directory, no_prompts, update, once, profile, migrate_pytest)


# nf-core subworkflows list subcommands
@subworkflows.group("list")
@click.pass_context
def subworkflows_list(ctx):
    """
    List subworkflows in a local pipeline or remote repository.
    """
    pass


# nf-core subworkflows list remote
@subworkflows_list.command("remote")
@click.pass_context
@click.argument("keywords", required=False, nargs=-1, metavar="<filter keywords>")
@click.option("-j", "--json", is_flag=True, help="Print as JSON to stdout")
def command_subworkflows_list_remote(ctx, keywords, json):
    """
    List subworkflows in a remote GitHub repo [dim i](e.g [link=https://github.com/nf-core/modules]nf-core/modules[/])[/].
    """
    subworkflows_list_remote(ctx, keywords, json)


# nf-core subworkflows list local
@subworkflows_list.command("local")
@click.pass_context
@click.argument("keywords", required=False, nargs=-1, metavar="<filter keywords>")
@click.option("-j", "--json", is_flag=True, help="Print as JSON to stdout")
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    help=r"Pipeline directory. [dim]\[default: Current working directory][/]",
)
def command_subworkflows_list_local(ctx, keywords, json, directory):  # pylint: disable=redefined-builtin
    """
    List subworkflows installed locally in a pipeline
    """
    subworkflows_list_local(ctx, keywords, json, directory)


# nf-core subworkflows lint
@subworkflows.command("lint")
@click.pass_context
@click.argument("subworkflow", type=str, callback=normalize_case, required=False, metavar="subworkflow name")
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    metavar="<pipeline/modules directory>",
)
@click.option(
    "-r",
    "--registry",
    type=str,
    metavar="<registry>",
    default=None,
    help="Registry to use for containers. If not specified it will use docker.registry value in the nextflow.config file",
)
@click.option(
    "-k",
    "--key",
    type=str,
    metavar="<test>",
    multiple=True,
    help="Run only these lint tests",
)
@click.option("-a", "--all", is_flag=True, help="Run on all subworkflows")
@click.option("-w", "--fail-warned", is_flag=True, help="Convert warn tests to failures")
@click.option("--local", is_flag=True, help="Run additional lint tests for local subworkflows")
@click.option("--passed", is_flag=True, help="Show passed tests")
@click.option(
    "--sort-by",
    type=click.Choice(["subworkflow", "test"]),
    default="test",
    help="Sort lint output by subworkflow or test name.",
    show_default=True,
)
@click.option("--fix", is_flag=True, help="Fix all linting tests if possible.")
def command_subworkflows_lint(
    ctx, subworkflow, directory, registry, key, all, fail_warned, local, passed, sort_by, fix
):
    """
    Lint one or more subworkflows in a directory.
    """
    subworkflows_lint(ctx, subworkflow, directory, registry, key, all, fail_warned, local, passed, sort_by, fix)


# nf-core subworkflows info
@subworkflows.command("info")
@click.pass_context
@click.argument("subworkflow", type=str, callback=normalize_case, required=False, metavar="subworkflow name")
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    help=r"Pipeline directory. [dim]\[default: Current working directory][/]",
)
def command_subworkflows_info(ctx, subworkflow, directory):
    """
    Show developer usage information about a given subworkflow.
    """
    subworkflows_info(ctx, subworkflow, directory)


# nf-core subworkflows install
@subworkflows.command("install")
@click.pass_context
@click.argument("subworkflow", type=str, callback=normalize_case, required=False, metavar="subworkflow name")
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    help=r"Pipeline directory. [dim]\[default: current working directory][/]",
)
@click.option(
    "-p",
    "--prompt",
    is_flag=True,
    default=False,
    help="Prompt for the version of the subworkflow",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force reinstallation of subworkflow if it already exists",
)
@click.option(
    "-s",
    "--sha",
    type=str,
    metavar="<commit sha>",
    help="Install subworkflow at commit SHA",
)
def command_subworkflows_install(ctx, subworkflow, directory, prompt, force, sha):
    """
    Install DSL2 subworkflow within a pipeline.
    """
    subworkflows_install(ctx, subworkflow, directory, prompt, force, sha)


# nf-core subworkflows patch
@subworkflows.command("patch")
@click.pass_context
@click.argument("tool", type=str, required=False, metavar="<tool> or <tool/subtool>")
@click.option(
    "-d",
    "--dir",
    type=click.Path(exists=True),
    default=".",
    help=r"Pipeline directory. [dim]\[default: current working directory][/]",
)
@click.option("-r", "--remove", is_flag=True, default=False, help="Remove an existent patch file and regenerate it.")
def subworkflows_patch(ctx, tool, dir, remove):
    """
    Create a patch file for minor changes in a subworkflow

    Checks if a subworkflow has been modified locally and creates a patch file
    describing how the module has changed from the remote version
    """
    from nf_core.subworkflows import SubworkflowPatch

    try:
        subworkflow_patch = SubworkflowPatch(
            dir,
            ctx.obj["modules_repo_url"],
            ctx.obj["modules_repo_branch"],
            ctx.obj["modules_repo_no_pull"],
        )
        if remove:
            subworkflow_patch.remove(tool)
        else:
            subworkflow_patch.patch(tool)
    except (UserWarning, LookupError) as e:
        log.error(e)
        sys.exit(1)


# nf-core subworkflows remove
@subworkflows.command("remove")
@click.pass_context
@click.argument("subworkflow", type=str, callback=normalize_case, required=False, metavar="subworkflow name")
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    help=r"Pipeline directory. [dim]\[default: current working directory][/]",
)
def command_subworkflows_remove(ctx, directory, subworkflow):
    """
    Remove a subworkflow from a pipeline.
    """
    subworkflows_remove(ctx, directory, subworkflow)


# nf-core subworkflows update
@subworkflows.command("update")
@click.pass_context
@click.argument("subworkflow", type=str, callback=normalize_case, required=False, metavar="subworkflow name")
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    help=r"Pipeline directory. [dim]\[default: current working directory][/]",
)
@click.option("-f", "--force", is_flag=True, default=False, help="Force update of subworkflow")
@click.option(
    "-p",
    "--prompt",
    is_flag=True,
    default=False,
    help="Prompt for the version of the subworkflow",
)
@click.option(
    "-s",
    "--sha",
    type=str,
    metavar="<commit sha>",
    help="Install subworkflow at commit SHA",
)
@click.option(
    "-l",
    "--limit-output",
    "limit_output",
    is_flag=True,
    default=False,
    help="Limit output to only the difference in main.nf",
)
@click.option(
    "-a",
    "--all",
    "install_all",
    is_flag=True,
    default=False,
    help="Update all subworkflow installed in pipeline",
)
@click.option(
    "-x/-y",
    "--preview/--no-preview",
    is_flag=True,
    default=None,
    help="Preview / no preview of changes before applying",
)
@click.option(
    "-D",
    "--save-diff",
    type=str,
    metavar="<filename>",
    default=None,
    help="Save diffs to a file instead of updating in place",
)
@click.option(
    "-u",
    "--update-deps",
    is_flag=True,
    default=False,
    help="Automatically update all linked modules and subworkflows without asking for confirmation",
)
def command_subworkflows_update(
    ctx,
    subworkflow,
    directory,
    force,
    prompt,
    sha,
    install_all,
    preview,
    save_diff,
    update_deps,
    limit_output,
):
    """
    Update DSL2 subworkflow within a pipeline.
    """
    subworkflows_update(
        ctx, subworkflow, directory, force, prompt, sha, install_all, preview, save_diff, update_deps, limit_output
    )


## DEPRECATED commands since v3.0.0


# nf-core schema subcommands (deprecated)
@nf_core_cli.group(deprecated=True, hidden=True)
def schema():
    """
    Use `nf-core pipelines schema <command>` instead.
    """
    pass


# nf-core schema validate (deprecated)
@schema.command("validate", deprecated=True)
@click.argument("pipeline", required=True, metavar="<pipeline name>")
@click.argument("params", type=click.Path(exists=True), required=True, metavar="<JSON params file>")
def command_schema_validate(pipeline, params):
    """
    Use `nf-core pipelines schema validate` instead.
    """
    log.warning(
        "The `[magenta]nf-core schema validate[/]` command is deprecated. Use `[magenta]nf-core pipelines schema validate[/]` instead."
    )
    pipelines_schema_validate(pipeline, params)


# nf-core schema build (deprecated)
@schema.command("build", deprecated=True)
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    help=r"Pipeline directory. [dim]\[default: current working directory][/]",
)
@click.option(
    "--no-prompts",
    is_flag=True,
    help="Do not confirm changes, just update parameters and exit",
)
@click.option(
    "--web-only",
    is_flag=True,
    help="Skip building using Nextflow config, just launch the web tool",
)
@click.option(
    "--url",
    type=str,
    default="https://oldsite.nf-co.re/pipeline_schema_builder",
    help="Customise the builder URL (for development work)",
)
def command_schema_build(directory, no_prompts, web_only, url):
    """
    Use `nf-core pipelines schema build` instead.
    """
    log.warning(
        "The `[magenta]nf-core schema build[/]` command is deprecated. Use `[magenta]nf-core pipelines schema build[/]` instead."
    )
    pipelines_schema_build(directory, no_prompts, web_only, url)


# nf-core schema lint (deprecated)
@schema.command("lint", deprecated=True)
@click.argument(
    "schema_path",
    type=click.Path(exists=True),
    default="nextflow_schema.json",
    metavar="<pipeline schema>",
)
def command_schema_lint(schema_path):
    """
    Use `nf-core pipelines schema lint` instead.
    """
    log.warning(
        "The `[magenta]nf-core schema lint[/]` command is deprecated. Use `[magenta]nf-core pipelines schema lint[/]` instead."
    )
    pipelines_schema_lint(schema_path)


# nf-core schema docs (deprecated)
@schema.command("docs", deprecated=True)
@click.argument(
    "schema_path",
    type=click.Path(exists=True),
    default="nextflow_schema.json",
    required=False,
    metavar="<pipeline schema>",
)
@click.option(
    "-o",
    "--output",
    type=str,
    metavar="<filename>",
    help="Output filename. Defaults to standard out.",
)
@click.option(
    "-x",
    "--format",
    type=click.Choice(["markdown", "html"]),
    default="markdown",
    help="Format to output docs in.",
)
@click.option("-f", "--force", is_flag=True, default=False, help="Overwrite existing files")
@click.option(
    "-c",
    "--columns",
    type=str,
    metavar="<columns_list>",
    help="CSV list of columns to include in the parameter tables (parameter,description,type,default,required,hidden)",
    default="parameter,description,type,default,required,hidden",
)
def command_schema_docs(schema_path, output, format, force, columns):
    """
    Use `nf-core pipelines schema docs` instead.
    """
    log.warning(
        "The `[magenta]nf-core schema docs[/]` command is deprecated. Use `[magenta]nf-core pipelines schema docs[/]` instead."
    )
    pipelines_schema_docs(schema_path, output, format, force, columns)


# nf-core create-logo (deprecated)
@nf_core_cli.command("create-logo", deprecated=True, hidden=True)
@click.argument("logo-text", metavar="<logo_text>")
@click.option("-d", "--dir", "directory", type=click.Path(), default=".", help="Directory to save the logo in.")
@click.option(
    "-n",
    "--name",
    type=str,
    help="Name of the output file (with or without '.png' suffix).",
)
@click.option(
    "--theme",
    type=click.Choice(["light", "dark"]),
    default="light",
    help="Theme for the logo.",
    show_default=True,
)
@click.option(
    "--width",
    type=int,
    default=2300,
    help="Width of the logo in pixels.",
    show_default=True,
)
@click.option(
    "--format",
    type=click.Choice(["png", "svg"]),
    default="png",
    help="Image format of the logo, either PNG or SVG.",
    show_default=True,
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Overwrite any files if they already exist",
)
def command_create_logo(logo_text, directory, name, theme, width, format, force):
    """
    Use `nf-core pipelines create-logo` instead.
    """
    log.warning(
        "The `[magenta]nf-core create-logo[/]` command is deprecated. Use `[magenta]nf-core pipeliness create-logo[/]` instead."
    )
    pipelines_create_logo(logo_text, directory, name, theme, width, format, force)


# nf-core sync (deprecated)
@nf_core_cli.command("sync", hidden=True, deprecated=True)
@click.pass_context
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    help=r"Pipeline directory. [dim]\[default: current working directory][/]",
)
@click.option(
    "-b",
    "--from-branch",
    type=str,
    help="The git branch to use to fetch workflow variables.",
)
@click.option(
    "-p",
    "--pull-request",
    is_flag=True,
    default=False,
    help="Make a GitHub pull-request with the changes.",
)
@click.option(
    "--force_pr",
    is_flag=True,
    default=False,
    help="Force the creation of a pull-request, even if there are no changes.",
)
@click.option("-g", "--github-repository", type=str, help="GitHub PR: target repository.")
@click.option("-u", "--username", type=str, help="GitHub PR: auth username.")
@click.option("-t", "--template-yaml", help="Pass a YAML file to customize the template")
def command_sync(ctx, directory, from_branch, pull_request, github_repository, username, template_yaml, force_pr):
    """
    Use `nf-core pipelines sync` instead.
    """
    log.warning(
        "The `[magenta]nf-core sync[/]` command is deprecated. Use `[magenta]nf-core pipelines sync[/]` instead."
    )
    pipelines_sync(ctx, directory, from_branch, pull_request, github_repository, username, template_yaml, force_pr)


# nf-core bump-version (deprecated)
@nf_core_cli.command("bump-version", hidden=True, deprecated=True)
@click.pass_context
@click.argument("new_version", default="")
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    help=r"Pipeline directory. [dim]\[default: current working directory][/]",
)
@click.option(
    "-n",
    "--nextflow",
    is_flag=True,
    default=False,
    help="Bump required nextflow version instead of pipeline version",
)
def command_bump_version(ctx, new_version, directory, nextflow):
    """
    Use `nf-core pipelines bump-version` instead.
    """
    log.warning(
        "The `[magenta]nf-core bump-version[/]` command is deprecated. Use `[magenta]nf-core pipelines bump-version[/]` instead."
    )
    pipelines_bump_version(ctx, new_version, directory, nextflow)


# nf-core list (deprecated)
@nf_core_cli.command("list", deprecated=True, hidden=True)
@click.argument("keywords", required=False, nargs=-1, metavar="<filter keywords>")
@click.option(
    "-s",
    "--sort",
    type=click.Choice(["release", "pulled", "name", "stars"]),
    default="release",
    help="How to sort listed pipelines",
)
@click.option("--json", is_flag=True, default=False, help="Print full output as JSON")
@click.option("--show-archived", is_flag=True, default=False, help="Print archived workflows")
@click.pass_context
def command_list(ctx, keywords, sort, json, show_archived):
    """
    DEPREUse `nf-core pipelines list` instead.CATED
    """
    log.warning(
        "The `[magenta]nf-core list[/]` command is deprecated. Use `[magenta]nf-core pipelines list[/]` instead."
    )
    pipelines_list(ctx, keywords, sort, json, show_archived)


# nf-core launch (deprecated)
@nf_core_cli.command("launch", deprecated=True, hidden=True)
@click.argument("pipeline", required=False, metavar="<pipeline name>")
@click.option("-r", "--revision", help="Release/branch/SHA of the project to run (if remote)")
@click.option("-i", "--id", help="ID for web-gui launch parameter set")
@click.option(
    "-c",
    "--command-only",
    is_flag=True,
    default=False,
    help="Create Nextflow command with params (no params file)",
)
@click.option(
    "-o",
    "--params-out",
    type=click.Path(),
    default=os.path.join(os.getcwd(), "nf-params.json"),
    help="Path to save run parameters file",
)
@click.option(
    "-p",
    "--params-in",
    type=click.Path(exists=True),
    help="Set of input run params to use from a previous run",
)
@click.option(
    "-a",
    "--save-all",
    is_flag=True,
    default=False,
    help="Save all parameters, even if unchanged from default",
)
@click.option(
    "-x",
    "--show-hidden",
    is_flag=True,
    default=False,
    help="Show hidden params which don't normally need changing",
)
@click.option(
    "-u",
    "--url",
    type=str,
    default="https://nf-co.re/launch",
    help="Customise the builder URL (for development work)",
)
@click.pass_context
def command_launch(
    ctx,
    pipeline,
    id,
    revision,
    command_only,
    params_in,
    params_out,
    save_all,
    show_hidden,
    url,
):
    """
    Use `nf-core pipelines launch` instead.
    """
    log.warning(
        "The `[magenta]nf-core launch[/]` command is deprecated. Use `[magenta]nf-core pipelines launch[/]` instead."
    )
    pipelines_launch(ctx, pipeline, id, revision, command_only, params_in, params_out, save_all, show_hidden, url)


# nf-core create-params-file (deprecated)
@nf_core_cli.command("create-params-file", deprecated=True, hidden=True)
@click.argument("pipeline", required=False, metavar="<pipeline name>")
@click.option("-r", "--revision", help="Release/branch/SHA of the pipeline (if remote)")
@click.option(
    "-o",
    "--output",
    type=str,
    default="nf-params.yml",
    metavar="<filename>",
    help="Output filename. Defaults to `nf-params.yml`.",
)
@click.option("-f", "--force", is_flag=True, default=False, help="Overwrite existing files")
@click.option(
    "-x",
    "--show-hidden",
    is_flag=True,
    default=False,
    help="Show hidden params which don't normally need changing",
)
def command_create_params_file(pipeline, revision, output, force, show_hidden):
    """
    Use `nf-core pipelines create-params-file` instead.
    """
    log.warning(
        "The `[magenta]nf-core create-params-file[/]` command is deprecated. Use `[magenta]nf-core pipelines create-params-file[/]` instead."
    )
    pipelines_create_params_file(pipeline, revision, output, force, show_hidden)


# nf-core download (deprecated)
@nf_core_cli.command("download", deprecated=True, hidden=True)
@click.argument("pipeline", required=False, metavar="<pipeline name>")
@click.option(
    "-r",
    "--revision",
    multiple=True,
    help="Pipeline release to download. Multiple invocations are possible, e.g. `-r 1.1 -r 1.2`",
)
@click.option("-o", "--outdir", type=str, help="Output directory")
@click.option(
    "-x",
    "--compress",
    type=click.Choice(["tar.gz", "tar.bz2", "zip", "none"]),
    help="Archive compression type",
)
@click.option("-f", "--force", is_flag=True, default=False, help="Overwrite existing files")
@click.option(
    "-t",
    "--tower",
    is_flag=True,
    default=False,
    hidden=True,
    help="Download for Seqera Platform. DEPRECATED: Please use `--platform` instead.",
)
@click.option(
    "--platform",
    is_flag=True,
    default=False,
    help="Download for Seqera Platform (formerly Nextflow Tower)",
)
@click.option(
    "-d",
    "--download-configuration",
    is_flag=True,
    default=False,
    help="Include configuration profiles in download. Not available with `--platform`",
)
@click.option(
    "--tag",
    multiple=True,
    help="Add custom alias tags to `--platform` downloads. For example, `--tag \"3.10=validated\"` adds the custom 'validated' tag to the 3.10 release.",
)
@click.option(
    "-s",
    "--container-system",
    type=click.Choice(["none", "singularity"]),
    help="Download container images of required software.",
)
@click.option(
    "-l",
    "--container-library",
    multiple=True,
    help="Container registry/library or mirror to pull images from.",
)
@click.option(
    "-u",
    "--container-cache-utilisation",
    type=click.Choice(["amend", "copy", "remote"]),
    help="Utilise a `singularity.cacheDir` in the download process, if applicable.",
)
@click.option(
    "-i",
    "--container-cache-index",
    type=str,
    help="List of images already available in a remote `singularity.cacheDir`.",
)
@click.option(
    "-p",
    "--parallel-downloads",
    type=int,
    default=4,
    help="Number of parallel image downloads",
)
@click.pass_context
def command_download(
    ctx,
    pipeline,
    revision,
    outdir,
    compress,
    force,
    tower,
    platform,
    download_configuration,
    tag,
    container_system,
    container_library,
    container_cache_utilisation,
    container_cache_index,
    parallel_downloads,
):
    """
    Use `nf-core pipelines download` instead.
    """
    log.warning(
        "The `[magenta]nf-core download[/]` command is deprecated. Use `[magenta]nf-core pipelines download[/]` instead."
    )
    pipelines_download(
        ctx,
        pipeline,
        revision,
        outdir,
        compress,
        force,
        platform or tower,
        download_configuration,
        tag,
        container_system,
        container_library,
        container_cache_utilisation,
        container_cache_index,
        parallel_downloads,
    )


# nf-core lint (deprecated)
@nf_core_cli.command("lint", hidden=True, deprecated=True)
@click.option(
    "-d",
    "--dir",
    "directory",
    type=click.Path(exists=True),
    default=".",
    help=r"Pipeline directory [dim]\[default: current working directory][/]",
)
@click.option(
    "--release",
    is_flag=True,
    default=Path(os.environ.get("GITHUB_REF", "").strip(" '\"")).parent.name in ["master", "main"]
    and os.environ.get("GITHUB_REPOSITORY", "").startswith("nf-core/")
    and not os.environ.get("GITHUB_REPOSITORY", "") == "nf-core/tools",
    help="Execute additional checks for release-ready workflows.",
)
@click.option(
    "-f",
    "--fix",
    type=str,
    metavar="<test>",
    multiple=True,
    help="Attempt to automatically fix specified lint test",
)
@click.option(
    "-k",
    "--key",
    type=str,
    metavar="<test>",
    multiple=True,
    help="Run only these lint tests",
)
@click.option("-p", "--show-passed", is_flag=True, help="Show passing tests on the command line")
@click.option("-i", "--fail-ignored", is_flag=True, help="Convert ignored tests to failures")
@click.option("-w", "--fail-warned", is_flag=True, help="Convert warn tests to failures")
@click.option(
    "--markdown",
    type=str,
    metavar="<filename>",
    help="File to write linting results to (Markdown)",
)
@click.option(
    "--json",
    type=str,
    metavar="<filename>",
    help="File to write linting results to (JSON)",
)
@click.option(
    "--sort-by",
    type=click.Choice(["module", "test"]),
    default="test",
    help="Sort lint output by module or test name.",
    show_default=True,
)
@click.pass_context
def command_lint(
    ctx,
    directory,
    release,
    fix,
    key,
    show_passed,
    fail_ignored,
    fail_warned,
    markdown,
    json,
    sort_by,
):
    """
    Use `nf-core pipelines lint` instead.
    """
    log.warning(
        "The `[magenta]nf-core lint[/]` command is deprecated. Use `[magenta]nf-core pipelines lint[/]` instead."
    )
    pipelines_lint(ctx, directory, release, fix, key, show_passed, fail_ignored, fail_warned, markdown, json, sort_by)


# nf-core create (deprecated)
@nf_core_cli.command("create", hidden=True, deprecated=True)
@click.option(
    "-n",
    "--name",
    type=str,
    help="The name of your new pipeline",
)
@click.option("-d", "--description", type=str, help="A short description of your pipeline")
@click.option("-a", "--author", type=str, help="Name of the main author(s)")
@click.option("--version", type=str, default="1.0.0dev", help="The initial version number to use")
@click.option("-f", "--force", is_flag=True, default=False, help="Overwrite output directory if it already exists")
@click.option("-o", "--outdir", help="Output directory for new pipeline (default: pipeline name)")
@click.option("-t", "--template-yaml", help="Pass a YAML file to customize the template")
@click.option("--plain", is_flag=True, help="Use the standard nf-core template")
@click.option(
    "--organisation",
    type=str,
    default="nf-core",
    help="The name of the GitHub organisation where the pipeline will be hosted (default: nf-core)",
)
@click.pass_context
def command_create(ctx, name, description, author, version, force, outdir, template_yaml, plain, organisation):
    """
    Use `nf-core pipelines create` instead.
    """
    log.warning(
        "The `[magenta]nf-core create[/]` command is deprecated. Use `[magenta]nf-core pipelines create[/]` instead."
    )
    pipelines_create(ctx, name, description, author, version, force, outdir, template_yaml, organisation)


# Main script is being run - launch the CLI
if __name__ == "__main__":
    run_nf_core()



import os
import tempfile
from pathlib import Path
from unittest import mock

import pytest

import nf_core.components.components_command
import nf_core.components.patch
import nf_core.modules.modules_json
import nf_core.modules.patch
import nf_core.modules.update

from ..test_modules import TestModules
from ..utils import GITLAB_URL

"""
Test the 'nf-core modules patch' command

Uses a branch (patch-tester) in the GitLab nf-core/modules-test repo when
testing if the update commands works correctly with patch files
"""

ORG_SHA = "3dc7c14d29af40f1a0871a675364e437559d97a8"
CORRECT_SHA = "63e780200600e340365b669f9c673b670764c569"
SUCCEED_SHA = "0d0515c3f11266e1314e129bec3e308f804c8dc7"
FAIL_SHA = "cb64a5c1ef85619b89ab99dec2e9097fe84e1dc8"
BISMARK_ALIGN = "bismark/align"
REPO_NAME = "nf-core-test"
PATCH_BRANCH = "patch-tester"
REPO_URL = "https://gitlab.com/nf-core/modules-test.git"


def setup_patch(pipeline_dir, modify_module):
    install_obj = nf_core.modules.install.ModuleInstall(
        pipeline_dir, prompt=False, force=False, remote_url=GITLAB_URL, branch=PATCH_BRANCH, sha=ORG_SHA
    )

    # Install the module
    install_obj.install(BISMARK_ALIGN)

    if modify_module:
        # Modify the module
        module_path = Path(pipeline_dir, "modules", REPO_NAME, BISMARK_ALIGN)
        modify_main_nf(module_path / "main.nf")


def modify_main_nf(path):
    """Modify a file to test patch creation"""
    with open(path) as fh:
        lines = fh.readlines()
    # We want a patch file that looks something like:
    # -    tuple val(meta), path(reads)
    # -    path index
    # +    tuple val(meta), path(reads), path(index)
    for line_index in range(len(lines)):
        if lines[line_index] == "    tuple val(meta), path(reads)\n":
            lines[line_index] = "    tuple val(meta), path(reads), path(index)\n"
        elif lines[line_index] == "    path index\n":
            to_pop = line_index
    lines.pop(to_pop)
    with open(path, "w") as fh:
        fh.writelines(lines)


class TestModulesCreate(TestModules):
    def test_create_patch_no_change(self):
        """Test creating a patch when there is no change to the module"""
        setup_patch(self.pipeline_dir, False)

        # Try creating a patch file
        patch_obj = nf_core.modules.patch.ModulePatch(self.pipeline_dir, GITLAB_URL, PATCH_BRANCH)
        with pytest.raises(UserWarning):
            patch_obj.patch(BISMARK_ALIGN)

        module_path = Path(self.pipeline_dir, "modules", REPO_NAME, BISMARK_ALIGN)

        # Check that no patch file has been added to the directory
        assert not (module_path / "bismark-align.diff").exists()

        # Check the 'modules.json' contains no patch file for the module
        modules_json_obj = nf_core.modules.modules_json.ModulesJson(self.pipeline_dir)
        assert modules_json_obj.get_patch_fn("modules", BISMARK_ALIGN, REPO_URL, REPO_NAME) is None

    def test_create_patch_change(self):
        """Test creating a patch when there is a change to the module"""
        setup_patch(self.pipeline_dir, True)

        # Try creating a patch file
        patch_obj = nf_core.modules.patch.ModulePatch(self.pipeline_dir, GITLAB_URL, PATCH_BRANCH)
        patch_obj.patch(BISMARK_ALIGN)

        module_path = Path(self.pipeline_dir, "modules", REPO_NAME, BISMARK_ALIGN)

        patch_fn = f"{'-'.join(BISMARK_ALIGN.split('/'))}.diff"
        # Check that a patch file with the correct name has been created
        assert (module_path / patch_fn).exists()

        # Check the 'modules.json' contains a patch file for the module
        modules_json_obj = nf_core.modules.modules_json.ModulesJson(self.pipeline_dir)
        assert modules_json_obj.get_patch_fn("modules", BISMARK_ALIGN, REPO_URL, REPO_NAME) == Path(
            "modules", REPO_NAME, BISMARK_ALIGN, patch_fn
        )

        # Check that the correct lines are in the patch file
        with open(module_path / patch_fn) as fh:
            patch_lines = fh.readlines()
        module_relpath = module_path.relative_to(self.pipeline_dir)
        assert f"--- {module_relpath / 'main.nf'}\n" in patch_lines, module_relpath / "main.nf"
        assert f"+++ {module_relpath / 'main.nf'}\n" in patch_lines
        assert "-    tuple val(meta), path(reads)\n" in patch_lines
        assert "-    path index\n" in patch_lines
        assert "+    tuple val(meta), path(reads), path(index)\n" in patch_lines

    def test_create_patch_try_apply_successful(self):
        """
        Test creating a patch file and applying it to a new version of the the files
        """

        setup_patch(self.pipeline_dir, True)
        module_relpath = Path("modules", REPO_NAME, BISMARK_ALIGN)
        module_path = Path(self.pipeline_dir, module_relpath)

        # Try creating a patch file
        patch_obj = nf_core.modules.patch.ModulePatch(self.pipeline_dir, GITLAB_URL, PATCH_BRANCH)
        patch_obj.patch(BISMARK_ALIGN)

        patch_fn = f"{'-'.join(BISMARK_ALIGN.split('/'))}.diff"
        # Check that a patch file with the correct name has been created
        assert (module_path / patch_fn).exists()

        # Check the 'modules.json' contains a patch file for the module
        modules_json_obj = nf_core.modules.modules_json.ModulesJson(self.pipeline_dir)
        assert modules_json_obj.get_patch_fn("modules", BISMARK_ALIGN, REPO_URL, REPO_NAME) == Path(
            "modules", REPO_NAME, BISMARK_ALIGN, patch_fn
        )

        update_obj = nf_core.modules.update.ModuleUpdate(
            self.pipeline_dir, sha=SUCCEED_SHA, remote_url=GITLAB_URL, branch=PATCH_BRANCH
        )
        # Install the new files
        install_dir = Path(tempfile.mkdtemp())
        update_obj.install_component_files(BISMARK_ALIGN, SUCCEED_SHA, update_obj.modules_repo, install_dir)

        # Try applying the patch
        module_install_dir = install_dir / BISMARK_ALIGN
        patch_relpath = module_relpath / patch_fn
        assert (
            update_obj.try_apply_patch(BISMARK_ALIGN, REPO_NAME, patch_relpath, module_path, module_install_dir) is True
        )

        # Move the files from the temporary directory
        update_obj.move_files_from_tmp_dir(BISMARK_ALIGN, install_dir, REPO_NAME, SUCCEED_SHA)

        # Check that a patch file with the correct name has been created
        assert (module_path / patch_fn).exists()

        # Check the 'modules.json' contains a patch file for the module
        modules_json_obj = nf_core.modules.modules_json.ModulesJson(self.pipeline_dir)
        assert modules_json_obj.get_patch_fn("modules", BISMARK_ALIGN, REPO_URL, REPO_NAME) == Path(
            "modules", REPO_NAME, BISMARK_ALIGN, patch_fn
        )

        # Check that the correct lines are in the patch file
        with open(module_path / patch_fn) as fh:
            patch_lines = fh.readlines()
        module_relpath = module_path.relative_to(self.pipeline_dir)
        assert f"--- {module_relpath / 'main.nf'}\n" in patch_lines
        assert f"+++ {module_relpath / 'main.nf'}\n" in patch_lines
        assert "-    tuple val(meta), path(reads)\n" in patch_lines
        assert "-    path index\n" in patch_lines
        assert "+    tuple val(meta), path(reads), path(index)\n" in patch_lines

        # Check that 'main.nf' is updated correctly
        with open(module_path / "main.nf") as fh:
            main_nf_lines = fh.readlines()
        # These lines should have been removed by the patch
        assert "    tuple val(meta), path(reads)\n" not in main_nf_lines
        assert "    path index\n" not in main_nf_lines
        # This line should have been added
        assert "    tuple val(meta), path(reads), path(index)\n" in main_nf_lines

    def test_create_patch_try_apply_failed(self):
        """
        Test creating a patch file and applying it to a new version of the the files
        """

        setup_patch(self.pipeline_dir, True)
        module_relpath = Path("modules", REPO_NAME, BISMARK_ALIGN)
        module_path = Path(self.pipeline_dir, module_relpath)

        # Try creating a patch file
        patch_obj = nf_core.modules.patch.ModulePatch(self.pipeline_dir, GITLAB_URL, PATCH_BRANCH)
        patch_obj.patch(BISMARK_ALIGN)

        patch_fn = f"{'-'.join(BISMARK_ALIGN.split('/'))}.diff"
        # Check that a patch file with the correct name has been created
        assert (module_path / patch_fn).exists()

        # Check the 'modules.json' contains a patch file for the module
        modules_json_obj = nf_core.modules.modules_json.ModulesJson(self.pipeline_dir)
        assert modules_json_obj.get_patch_fn("modules", BISMARK_ALIGN, REPO_URL, REPO_NAME) == Path(
            "modules", REPO_NAME, BISMARK_ALIGN, patch_fn
        )

        update_obj = nf_core.modules.update.ModuleUpdate(
            self.pipeline_dir, sha=FAIL_SHA, remote_url=GITLAB_URL, branch=PATCH_BRANCH
        )
        # Install the new files
        install_dir = Path(tempfile.mkdtemp())
        update_obj.install_component_files(BISMARK_ALIGN, FAIL_SHA, update_obj.modules_repo, install_dir)

        # Try applying the patch
        module_install_dir = install_dir / BISMARK_ALIGN
        patch_path = module_relpath / patch_fn
        assert (
            update_obj.try_apply_patch(BISMARK_ALIGN, REPO_NAME, patch_path, module_path, module_install_dir) is False
        )

    def test_create_patch_update_success(self):
        """
        Test creating a patch file and the updating the module

        Should have the same effect as 'test_create_patch_try_apply_successful'
        but uses higher level api
        """

        setup_patch(self.pipeline_dir, True)
        module_path = Path(self.pipeline_dir, "modules", REPO_NAME, BISMARK_ALIGN)

        # Try creating a patch file
        patch_obj = nf_core.modules.patch.ModulePatch(self.pipeline_dir, GITLAB_URL, PATCH_BRANCH)
        patch_obj.patch(BISMARK_ALIGN)

        patch_fn = f"{'-'.join(BISMARK_ALIGN.split('/'))}.diff"
        # Check that a patch file with the correct name has been created
        assert (module_path / patch_fn).exists()

        # Check the 'modules.json' contains a patch file for the module
        modules_json_obj = nf_core.modules.modules_json.ModulesJson(self.pipeline_dir)
        assert modules_json_obj.get_patch_fn("modules", BISMARK_ALIGN, GITLAB_URL, REPO_NAME) == Path(
            "modules", REPO_NAME, BISMARK_ALIGN, patch_fn
        )

        # Update the module
        update_obj = nf_core.modules.update.ModuleUpdate(
            self.pipeline_dir,
            sha=SUCCEED_SHA,
            show_diff=False,
            update_deps=True,
            remote_url=GITLAB_URL,
            branch=PATCH_BRANCH,
        )
        assert update_obj.update(BISMARK_ALIGN)

        # Check that a patch file with the correct name has been created
        assert (module_path / patch_fn).exists()

        # Check the 'modules.json' contains a patch file for the module
        modules_json_obj = nf_core.modules.modules_json.ModulesJson(self.pipeline_dir)
        assert modules_json_obj.get_patch_fn("modules", BISMARK_ALIGN, GITLAB_URL, REPO_NAME) == Path(
            "modules", REPO_NAME, BISMARK_ALIGN, patch_fn
        ), modules_json_obj.get_patch_fn("modules", BISMARK_ALIGN, GITLAB_URL, REPO_NAME)

        # Check that the correct lines are in the patch file
        with open(module_path / patch_fn) as fh:
            patch_lines = fh.readlines()
        module_relpath = module_path.relative_to(self.pipeline_dir)
        assert f"--- {module_relpath / 'main.nf'}\n" in patch_lines
        assert f"+++ {module_relpath / 'main.nf'}\n" in patch_lines
        assert "-    tuple val(meta), path(reads)\n" in patch_lines
        assert "-    path index\n" in patch_lines
        assert "+    tuple val(meta), path(reads), path(index)\n" in patch_lines

        # Check that 'main.nf' is updated correctly
        with open(module_path / "main.nf") as fh:
            main_nf_lines = fh.readlines()
        # These lines should have been removed by the patch
        assert "    tuple val(meta), path(reads)\n" not in main_nf_lines
        assert "    path index\n" not in main_nf_lines
        # This line should have been added
        assert "    tuple val(meta), path(reads), path(index)\n" in main_nf_lines

    def test_create_patch_update_fail(self):
        """
        Test creating a patch file and updating a module when there is a diff conflict
        """

        setup_patch(self.pipeline_dir, True)
        module_path = Path(self.pipeline_dir, "modules", REPO_NAME, BISMARK_ALIGN)

        # Try creating a patch file
        patch_obj = nf_core.modules.patch.ModulePatch(self.pipeline_dir, GITLAB_URL, PATCH_BRANCH)
        patch_obj.patch(BISMARK_ALIGN)

        patch_fn = f"{'-'.join(BISMARK_ALIGN.split('/'))}.diff"
        # Check that a patch file with the correct name has been created
        assert (module_path / patch_fn).exists()

        # Check the 'modules.json' contains a patch file for the module
        modules_json_obj = nf_core.modules.modules_json.ModulesJson(self.pipeline_dir)
        assert modules_json_obj.get_patch_fn("modules", BISMARK_ALIGN, REPO_URL, REPO_NAME) == Path(
            "modules", REPO_NAME, BISMARK_ALIGN, patch_fn
        )

        # Save the file contents for downstream comparison
        with open(module_path / patch_fn) as fh:
            patch_contents = fh.read()

        update_obj = nf_core.modules.update.ModuleUpdate(
            self.pipeline_dir,
            sha=FAIL_SHA,
            show_diff=False,
            update_deps=True,
            remote_url=GITLAB_URL,
            branch=PATCH_BRANCH,
        )
        update_obj.update(BISMARK_ALIGN)

        # Check that the installed files have not been affected by the attempted patch
        temp_dir = Path(tempfile.mkdtemp())
        nf_core.components.components_command.ComponentCommand(
            "modules", self.pipeline_dir, GITLAB_URL, PATCH_BRANCH
        ).install_component_files(BISMARK_ALIGN, FAIL_SHA, update_obj.modules_repo, temp_dir)

        temp_module_dir = temp_dir / BISMARK_ALIGN
        for file in os.listdir(temp_module_dir):
            assert file in os.listdir(module_path)
            with open(module_path / file) as fh:
                installed = fh.read()
            with open(temp_module_dir / file) as fh:
                shouldbe = fh.read()
            assert installed == shouldbe

        # Check that the patch file is unaffected
        with open(module_path / patch_fn) as fh:
            new_patch_contents = fh.read()
        assert patch_contents == new_patch_contents

    def test_remove_patch(self):
        """Test creating a patch when there is no change to the module"""
        setup_patch(self.pipeline_dir, True)

        # Try creating a patch file
        patch_obj = nf_core.modules.patch.ModulePatch(self.pipeline_dir, GITLAB_URL, PATCH_BRANCH)
        patch_obj.patch(BISMARK_ALIGN)

        module_path = Path(self.pipeline_dir, "modules", REPO_NAME, BISMARK_ALIGN)

        # Check that a patch file with the correct name has been created
        patch_fn = f"{'-'.join(BISMARK_ALIGN.split('/'))}.diff"
        assert (module_path / patch_fn).exists()

        # Check the 'modules.json' contains a patch file for the module
        modules_json_obj = nf_core.modules.modules_json.ModulesJson(self.pipeline_dir)
        assert modules_json_obj.get_patch_fn("modules", BISMARK_ALIGN, REPO_URL, REPO_NAME) == Path(
            "modules", REPO_NAME, BISMARK_ALIGN, patch_fn
        )

        with mock.patch.object(nf_core.components.patch.questionary, "confirm") as mock_questionary:
            mock_questionary.unsafe_ask.return_value = True
            patch_obj.remove(BISMARK_ALIGN)
        # Check that the diff file has been removed
        assert not (module_path / patch_fn).exists()

        # Check that the 'modules.json' entry has been removed
        modules_json_obj = nf_core.modules.modules_json.ModulesJson(self.pipeline_dir)
        assert modules_json_obj.get_patch_fn("modules", BISMARK_ALIGN, REPO_URL, REPO_NAME) is None



import logging
from pathlib import Path
from typing import Dict, List, Union

log = logging.getLogger(__name__)


def files_exist(self) -> Dict[str, List[str]]:
    """Checks a given pipeline directory for required files.

    Iterates through the pipeline's directory content and checks that specified
    files are either present or absent, as required.

    .. note::
        This test raises an ``AssertionError`` if neither ``nextflow.config`` or ``main.nf`` are found.
        If these files are not found then this cannot be a Nextflow pipeline and something has gone badly wrong.
        All lint tests are stopped immediately with a critical error message.

    Files that *must* be present:

    .. code-block:: bash

        .gitattributes
        .gitignore
        .nf-core.yml
        .editorconfig
        .prettierignore
        .prettierrc.yml
        .github/.dockstore.yml
        .github/CONTRIBUTING.md
        .github/ISSUE_TEMPLATE/bug_report.yml
        .github/ISSUE_TEMPLATE/config.yml
        .github/ISSUE_TEMPLATE/feature_request.yml
        .github/PULL_REQUEST_TEMPLATE.md
        .github/workflows/branch.yml
        .github/workflows/ci.yml
        .github/workflows/linting_comment.yml
        .github/workflows/linting.yml
        [LICENSE, LICENSE.md, LICENCE, LICENCE.md]  # NB: British / American spelling
        assets/email_template.html
        assets/email_template.txt
        assets/nf-core-PIPELINE_logo_light.png
        assets/sendmail_template.txt
        conf/modules.config
        conf/test.config
        conf/test_full.config
        CHANGELOG.md
        CITATIONS.md
        CODE_OF_CONDUCT.md
        docs/images/nf-core-PIPELINE_logo_light.png
        docs/images/nf-core-PIPELINE_logo_dark.png
        docs/output.md
        docs/README.md
        docs/usage.md
        nextflow_schema.json
        nextflow.config
        README.md

    Files that *should* be present:

    .. code-block:: bash

        main.nf
        assets/multiqc_config.yml
        conf/base.config
        conf/igenomes.config
        .github/workflows/awstest.yml
        .github/workflows/awsfulltest.yml
        ro-crate-metadata.json

    Files that *must not* be present, due to being renamed or removed in the template:

    .. code-block:: bash

        .github/ISSUE_TEMPLATE/bug_report.md
        .github/ISSUE_TEMPLATE/feature_request.md
        .github/workflows/push_dockerhub.yml
        .markdownlint.yml
        .nf-core.yaml  # NB: Should be yml, not yaml
        .yamllint.yml
        bin/markdown_to_html.r
        conf/aws.config
        docs/images/nf-core-PIPELINE_logo.png
        lib/Checks.groovy
        lib/Completion.groovy
        lib/NfcoreTemplate.groovy
        lib/Utils.groovy
        lib/Workflow.groovy
        lib/WorkflowMain.groovy
        lib/WorkflowPIPELINE.groovy
        lib/nfcore_external_java_deps.jar
        parameters.settings.json
        pipeline_template.yml # saving information in .nf-core.yml
        Singularity


    Files that *should not* be present:

    .. code-block:: bash

        .travis.yml

    .. tip:: You can configure the ``nf-core pipelines lint`` tests to ignore any of these checks by setting
            the ``files_exist`` key as follows in your ``.nf-core.yml`` config file. For example:

            .. code-block:: yaml

            lint:
                files_exist:
                    - assets/multiqc_config.yml
    """

    passed = []
    warned = []
    failed = []
    ignored = []

    # NB: Should all be files, not directories
    # List of lists. Passes if any of the files in the sublist are found.
    #: test autodoc
    try:
        _, short_name = self.nf_config["manifest.name"].strip("\"'").split("/")
    except ValueError:
        log.warning("Expected manifest.name to be in the format '<repo>/<pipeline>'. Will assume it is '<pipeline>'.")
        short_name = self.nf_config["manifest.name"].strip("\"'").split("/")

    files_fail = [
        [Path(".gitattributes")],
        [Path(".gitignore")],
        [Path(".nf-core.yml")],
        [Path(".editorconfig")],
        [Path(".prettierignore")],
        [Path(".prettierrc.yml")],
        [Path("CHANGELOG.md")],
        [Path("CITATIONS.md")],
        [Path("CODE_OF_CONDUCT.md")],
        [Path("LICENSE"), Path("LICENSE.md"), Path("LICENCE"), Path("LICENCE.md")],  # NB: British / American spelling
        [Path("nextflow_schema.json")],
        [Path("nextflow.config")],
        [Path("README.md")],
        [Path(".github", ".dockstore.yml")],
        [Path(".github", "CONTRIBUTING.md")],
        [Path(".github", "ISSUE_TEMPLATE", "bug_report.yml")],
        [Path(".github", "ISSUE_TEMPLATE", "config.yml")],
        [Path(".github", "ISSUE_TEMPLATE", "feature_request.yml")],
        [Path(".github", "PULL_REQUEST_TEMPLATE.md")],
        [Path(".github", "workflows", "branch.yml")],
        [Path(".github", "workflows", "ci.yml")],
        [Path(".github", "workflows", "linting_comment.yml")],
        [Path(".github", "workflows", "linting.yml")],
        [Path("assets", "email_template.html")],
        [Path("assets", "email_template.txt")],
        [Path("assets", "sendmail_template.txt")],
        [Path("assets", f"nf-core-{short_name}_logo_light.png")],
        [Path("conf", "modules.config")],
        [Path("conf", "test.config")],
        [Path("conf", "test_full.config")],
        [Path("docs", "images", f"nf-core-{short_name}_logo_light.png")],
        [Path("docs", "images", f"nf-core-{short_name}_logo_dark.png")],
        [Path("docs", "output.md")],
        [Path("docs", "README.md")],
        [Path("docs", "README.md")],
        [Path("docs", "usage.md")],
    ]

    files_warn = [
        [Path("main.nf")],
        [Path("assets", "multiqc_config.yml")],
        [Path("conf", "base.config")],
        [Path("conf", "igenomes.config")],
        [Path("conf", "igenomes_ignored.config")],
        [Path(".github", "workflows", "awstest.yml")],
        [Path(".github", "workflows", "awsfulltest.yml")],
        [Path("modules.json")],
        [Path("ro-crate-metadata.json")],
    ]

    # List of strings. Fails / warns if any of the strings exist.
    files_fail_ifexists = [
        Path(".github", "ISSUE_TEMPLATE", "bug_report.md"),
        Path(".github", "ISSUE_TEMPLATE", "feature_request.md"),
        Path(".github", "workflows", "push_dockerhub.yml"),
        Path(".markdownlint.yml"),
        Path(".nf-core.yaml"),  # yml not yaml
        Path(".yamllint.yml"),
        Path("bin", "markdown_to_html.r"),
        Path("conf", "aws.config"),
        Path("docs", "images", f"nf-core-{short_name}_logo.png"),
        Path("lib", "Checks.groovy"),
        Path("lib", "Completion.groovy"),
        Path("lib", "NfcoreTemplate.groovy"),
        Path("lib", "Utils.groovy"),
        Path("lib", "Workflow.groovy"),
        Path("lib", "WorkflowMain.groovy"),
        Path("lib", f"Workflow{short_name[0].upper()}{short_name[1:]}.groovy"),
        Path("parameters.settings.json"),
        Path("pipeline_template.yml"),  # saving information in .nf-core.yml
        Path("Singularity"),
        Path("lib", "nfcore_external_java_deps.jar"),
    ]
    files_warn_ifexists = [Path(".travis.yml")]

    files_hint = [
        [
            ["ro-crate-metadata.json"],
            ". Run `nf-core rocrate` to generate this file. Read more about RO-Crates in the [nf-core/tools docs](https://nf-co.re/tools#create-a-ro-crate-metadata-file).",
        ],
    ]
    # Remove files that should be ignored according to the linting config
    ignore_files = self.lint_config.get("files_exist", []) if self.lint_config is not None else []

    def pf(file_path: Union[str, Path]) -> Path:
        return Path(self.wf_path, file_path)

    # First - critical files. Check that this is actually a Nextflow pipeline
    if not pf("nextflow.config").is_file() and not pf("main.nf").is_file():
        failed.append("File not found: nextflow.config or main.nf")
        raise AssertionError("Neither nextflow.config or main.nf found! Is this a Nextflow pipeline?")

    # Files that cause an error if they don't exist
    for files in files_fail:
        if any([str(f) in ignore_files for f in files]):
            continue
        if any([pf(f).is_file() for f in files]):
            passed.append(f"File found: {self._wrap_quotes(files)}")
        else:
            failed.append(f"File not found: {self._wrap_quotes(files)}")

    # Files that cause a warning if they don't exist
    for files in files_warn:
        if any([str(f) in ignore_files for f in files]):
            continue
        if any([pf(f).is_file() for f in files]):
            passed.append(f"File found: {self._wrap_quotes(files)}")
        else:
            hint = ""
            for file_hint in files_hint:
                if file_hint[0] == files:
                    hint = str(file_hint[1])
            warned.append(f"File not found: {self._wrap_quotes(files)}{hint}")

    # Files that cause an error if they exist
    for file in files_fail_ifexists:
        if str(file) in ignore_files:
            continue
        if pf(file).is_file():
            failed.append(f"File must be removed: {self._wrap_quotes(file)}")
        else:
            passed.append(f"File not found check: {self._wrap_quotes(file)}")

    # Files that cause a warning if they exist
    for file in files_warn_ifexists:
        if str(file) in ignore_files:
            continue
        if pf(file).is_file():
            warned.append(f"File should be removed: {self._wrap_quotes(file)}")
        else:
            passed.append(f"File not found check: {self._wrap_quotes(file)}")

    # Files that are ignored
    for file in ignore_files:
        ignored.append(f"File is ignored: {self._wrap_quotes(file)}")

    return {"passed": passed, "warned": warned, "failed": failed, "ignored": ignored}



import shutil
from pathlib import Path
from unittest import mock

import pytest
import yaml
from git.repo import Repo

import nf_core.subworkflows
from tests.utils import GITLAB_SUBWORKFLOWS_ORG_PATH_BRANCH, GITLAB_URL

from ..test_subworkflows import TestSubworkflows


class TestSubworkflowsCreate(TestSubworkflows):
    def test_subworkflows_create_succeed(self):
        """Succeed at creating a subworkflow from the template inside a pipeline"""
        subworkflow_create = nf_core.subworkflows.SubworkflowCreate(
            self.pipeline_dir, "test_subworkflow_local", "@author", True
        )
        subworkflow_create.create()
        assert Path(self.pipeline_dir, "subworkflows", "local", "test_subworkflow_local.nf").exists()

    def test_subworkflows_create_fail_exists(self):
        """Fail at creating the same subworkflow twice"""
        subworkflow_create = nf_core.subworkflows.SubworkflowCreate(
            self.pipeline_dir, "test_subworkflow2", "@author", False
        )
        subworkflow_create.create()
        with pytest.raises(UserWarning) as excinfo:
            subworkflow_create.create()
        assert "Subworkflow file exists already" in str(excinfo.value)

    def test_subworkflows_create_nfcore_modules(self):
        """Create a subworkflow in nf-core/modules clone"""
        subworkflow_create = nf_core.subworkflows.SubworkflowCreate(
            self.nfcore_modules, "test_subworkflow", "@author", force=True
        )
        subworkflow_create.create()
        assert Path(self.nfcore_modules, "subworkflows", "nf-core", "test_subworkflow", "main.nf").exists()

        assert Path(
            self.nfcore_modules, "subworkflows", "nf-core", "test_subworkflow", "tests", "main.nf.test"
        ).exists()

    @mock.patch("rich.prompt.Confirm.ask")
    def test_subworkflows_migrate(self, mock_rich_ask):
        """Create a subworkflow with the --migrate-pytest option to convert pytest to nf-test"""
        pytest_dir = Path(self.nfcore_modules, "tests", "subworkflows", "nf-core", "bam_stats_samtools")
        subworkflow_dir = Path(self.nfcore_modules, "subworkflows", "nf-core", "bam_stats_samtools")

        # Clone modules repo with pytests
        shutil.rmtree(self.nfcore_modules)
        Repo.clone_from(GITLAB_URL, self.nfcore_modules, branch=GITLAB_SUBWORKFLOWS_ORG_PATH_BRANCH)
        with open(subworkflow_dir / "main.nf") as fh:
            old_main_nf = fh.read()
        with open(subworkflow_dir / "meta.yml") as fh:
            old_meta_yml = fh.read()

        # Create a subworkflow with --migrate-pytest
        mock_rich_ask.return_value = True
        subworkflow_create = nf_core.subworkflows.SubworkflowCreate(
            self.nfcore_modules, "bam_stats_samtools", migrate_pytest=True
        )
        subworkflow_create.create()

        with open(subworkflow_dir / "main.nf") as fh:
            new_main_nf = fh.read()
        with open(subworkflow_dir / "meta.yml") as fh:
            new_meta_yml = fh.read()
        nextflow_config = subworkflow_dir / "tests" / "nextflow.config"

        # Check that old files have been copied to the new module
        assert old_main_nf == new_main_nf
        assert old_meta_yml == new_meta_yml
        assert nextflow_config.is_file()

        # Check that pytest folder is deleted
        assert not pytest_dir.is_dir()

        # Check that pytest_modules.yml is updated
        with open(Path(self.nfcore_modules, "tests", "config", "pytest_modules.yml")) as fh:
            modules_yml = yaml.safe_load(fh)
        assert "subworkflows/bam_stats_samtools" not in modules_yml.keys()

    @mock.patch("rich.prompt.Confirm.ask")
    def test_subworkflows_migrate_no_delete(self, mock_rich_ask):
        """Create a subworkflow with the --migrate-pytest option to convert pytest to nf-test.
        Test that pytest directory is not deleted."""
        pytest_dir = Path(self.nfcore_modules, "tests", "subworkflows", "nf-core", "bam_stats_samtools")

        # Clone modules repo with pytests
        shutil.rmtree(self.nfcore_modules)
        Repo.clone_from(GITLAB_URL, self.nfcore_modules, branch=GITLAB_SUBWORKFLOWS_ORG_PATH_BRANCH)

        # Create a module with --migrate-pytest
        mock_rich_ask.return_value = False
        module_create = nf_core.subworkflows.SubworkflowCreate(
            self.nfcore_modules, "bam_stats_samtools", migrate_pytest=True
        )
        module_create.create()

        # Check that pytest folder is not deleted
        assert pytest_dir.is_dir()

        # Check that pytest_modules.yml is updated
        with open(Path(self.nfcore_modules, "tests", "config", "pytest_modules.yml")) as fh:
            modules_yml = yaml.safe_load(fh)
        assert "subworkflows/bam_stats_samtools" not in modules_yml.keys()



from pathlib import Path

from ruamel.yaml import YAML

import nf_core.pipelines.lint
from nf_core.utils import NFCoreYamlConfig

from ..test_lint import TestLint


class TestLintNfCoreYml(TestLint):
    def setUp(self) -> None:
        super().setUp()
        self.new_pipeline = self._make_pipeline_copy()
        self.nf_core_yml_path = Path(self.new_pipeline) / ".nf-core.yml"
        self.yaml = YAML()
        self.nf_core_yml: NFCoreYamlConfig = self.yaml.load(self.nf_core_yml_path)
        self.lint_obj = nf_core.pipelines.lint.PipelineLint(self.new_pipeline)

    def test_nfcore_yml_pass(self):
        """Lint test: nfcore_yml - PASS"""
        assert self.lint_obj._load()
        results = self.lint_obj.nfcore_yml()

        assert "Repository type in `.nf-core.yml` is valid" in str(results["passed"])
        assert "nf-core version in `.nf-core.yml` is set to the latest version" in str(results["passed"])
        assert len(results.get("warned", [])) == 0
        assert len(results.get("failed", [])) == 0
        assert len(results.get("ignored", [])) == 0

    def test_nfcore_yml_fail_repo_type(self):
        """Lint test: nfcore_yml - FAIL - repository type not set"""

        self.nf_core_yml["repository_type"] = "foo"
        self.yaml.dump(self.nf_core_yml, self.nf_core_yml_path)
        with self.assertRaises(AssertionError):
            self.lint_obj._load()

    def test_nfcore_yml_fail_nfcore_version(self):
        """Lint test: nfcore_yml - FAIL - nf-core version not set"""

        self.nf_core_yml["nf_core_version"] = "foo"
        self.yaml.dump(self.nf_core_yml, self.nf_core_yml_path)
        assert self.lint_obj._load()
        results = self.lint_obj.nfcore_yml()
        assert "nf-core version in `.nf-core.yml` is not set to the latest version." in str(results["warned"])
        assert len(results.get("failed", [])) == 0
        assert len(results.get("passed", [])) >= 0
        assert len(results.get("ignored", [])) == 0

    def test_nfcore_yml_nested_lint_config(self) -> None:
        """Lint test: nfcore_yml with nested lint config - PASS"""
        valid_yaml = """
        lint:
            files_unchanged:
                - .github/workflows/branch.yml
            # modules_config: False
            modules_config:
                    - fastqc
            # merge_markers: False
            merge_markers:
                    - docs/my_pdf.pdf
            # nextflow_config: False
            nextflow_config:
                - manifest.name
                - config_defaults:
                    - params.annotation_db
                    - params.multiqc_comment_headers
                    - params.custom_table_headers
            multiqc_config:
                - report_section_order
                - report_comment
            files_exist:
                - .github/CONTRIBUTING.md
                - CITATIONS.md
            # template_strings: False
            template_strings:
                    - docs/my_pdf.pdf
        """
        self.nf_core_yml["lint"] = self.yaml.load(valid_yaml)
        self.yaml.dump(self.nf_core_yml, self.nf_core_yml_path)

        assert self.lint_obj._load()
        results = self.lint_obj.nfcore_yml()
        assert len(results.get("failed", [])) == 0
        assert len(results.get("warned", [])) == 0
        assert len(results.get("ignored", [])) == 0

    def test_nfcore_yml_nested_lint_config_bool(self) -> None:
        """Lint test: nfcore_yml with nested lint config - PASS"""
        valid_yaml = """
        lint:
            files_unchanged:
                - .github/workflows/branch.yml
            modules_config: False
            # modules_config:
            #         - fastqc
            merge_markers: False
            # merge_markers:
            #         - docs/my_pdf.pdf
            # nextflow_config: False
            nextflow_config:
                - manifest.name
                - config_defaults:
                    - params.annotation_db
                    - params.multiqc_comment_headers
                    - params.custom_table_headers
            multiqc_config:
                - report_section_order
                - report_comment
            files_exist:
                - .github/CONTRIBUTING.md
                - CITATIONS.md
            template_strings: False
            # template_strings:
            #         - docs/my_pdf.pdf
        """
        self.nf_core_yml["lint"] = self.yaml.load(valid_yaml)
        self.yaml.dump(self.nf_core_yml, self.nf_core_yml_path)

        assert self.lint_obj._load()
        results = self.lint_obj.nfcore_yml()
        assert len(results.get("failed", [])) == 0
        assert len(results.get("warned", [])) == 0
        assert len(results.get("ignored", [])) == 0



from rich.console import Console

import nf_core.modules.info

from ..test_modules import TestModules
from ..utils import GITLAB_DEFAULT_BRANCH, GITLAB_URL


class TestModulesCreate(TestModules):
    def test_modules_info_remote(self):
        """Test getting info about a remote module"""
        mods_info = nf_core.modules.info.ModuleInfo(self.pipeline_dir, "fastqc")
        mods_info_output = mods_info.get_component_info()
        console = Console(record=True)
        console.print(mods_info_output)
        output = console.export_text()

        assert "Module: fastqc" in output
        assert "Inputs" in output
        assert "Outputs" in output

    def test_modules_info_remote_gitlab(self):
        """Test getting info about a module in the remote gitlab repo"""
        mods_info = nf_core.modules.info.ModuleInfo(
            self.pipeline_dir, "fastqc", remote_url=GITLAB_URL, branch=GITLAB_DEFAULT_BRANCH
        )
        mods_info_output = mods_info.get_component_info()
        console = Console(record=True)
        console.print(mods_info_output)
        output = console.export_text()

        assert "Module: fastqc" in output
        assert "Inputs" in output
        assert "Outputs" in output
        assert "--git-remote" in output

    def test_modules_info_local(self):
        """Test getting info about a locally installed module"""
        self.mods_install.install("trimgalore")
        mods_info = nf_core.modules.info.ModuleInfo(self.pipeline_dir, "trimgalore")
        mods_info_output = mods_info.get_component_info()
        console = Console(record=True)
        console.print(mods_info_output)
        output = console.export_text()

        assert "Module: trimgalore" in output
        assert "Inputs" in output
        assert "Outputs" in output
        assert "Location" in output

    def test_modules_info_in_modules_repo(self):
        """Test getting info about a module in the modules repo"""
        mods_info = nf_core.modules.info.ModuleInfo(self.nfcore_modules, "fastqc")
        mods_info.local = True
        mods_info_output = mods_info.get_component_info()
        console = Console(record=True)
        console.print(mods_info_output)
        output = console.export_text()

        assert "Module: fastqc" in output
        assert "Inputs" in output
        assert "Outputs" in output



"""A Textual app to create a pipeline."""

import logging

import click
from textual.app import App
from textual.widgets import Button, Switch

from nf_core.pipelines.create import utils
from nf_core.pipelines.create.basicdetails import BasicDetails
from nf_core.pipelines.create.custompipeline import CustomPipeline
from nf_core.pipelines.create.finaldetails import FinalDetails
from nf_core.pipelines.create.githubexit import GithubExit
from nf_core.pipelines.create.githubrepo import GithubRepo
from nf_core.pipelines.create.githubrepoquestion import GithubRepoQuestion
from nf_core.pipelines.create.loggingscreen import LoggingScreen
from nf_core.pipelines.create.nfcorepipeline import NfcorePipeline
from nf_core.pipelines.create.pipelinetype import ChoosePipelineType
from nf_core.pipelines.create.welcome import WelcomeScreen

log_handler = utils.CustomLogHandler(
    console=utils.LoggingConsole(classes="log_console"),
    rich_tracebacks=True,
    show_time=False,
    show_path=False,
    markup=True,
    tracebacks_suppress=[click],
)
logging.basicConfig(
    level="INFO",
    handlers=[log_handler],
    format="%(message)s",
)
log_handler.setLevel("INFO")


class PipelineCreateApp(App[utils.CreateConfig]):
    """A Textual app to manage stopwatches."""

    CSS_PATH = "create.tcss"
    TITLE = "nf-core pipelines create"
    SUB_TITLE = "Create a new pipeline with the nf-core pipeline template"
    BINDINGS = [
        ("d", "toggle_dark", "Toggle dark mode"),
        ("q", "quit", "Quit"),
        ("a", "toggle_all", "Toggle all"),
    ]
    SCREENS = {
        "welcome": WelcomeScreen(),
        "basic_details": BasicDetails(),
        "choose_type": ChoosePipelineType(),
        "type_custom": CustomPipeline(),
        "type_nfcore": NfcorePipeline(),
        "final_details": FinalDetails(),
        "logging": LoggingScreen(),
        "github_repo_question": GithubRepoQuestion(),
        "github_repo": GithubRepo(),
        "github_exit": GithubExit(),
    }

    # Initialise config as empty
    TEMPLATE_CONFIG = utils.CreateConfig()

    # Initialise pipeline type
    NFCORE_PIPELINE = True

    # Log handler
    LOG_HANDLER = log_handler
    # Logging state
    LOGGING_STATE = None

    # Template features
    template_features_yml = utils.load_features_yaml()

    def on_mount(self) -> None:
        self.push_screen("welcome")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Handle all button pressed events."""
        if event.button.id == "start":
            self.push_screen("choose_type")
        elif event.button.id == "type_nfcore":
            self.NFCORE_PIPELINE = True
            utils.NFCORE_PIPELINE_GLOBAL = True
            self.push_screen("basic_details")
        elif event.button.id == "type_custom":
            self.NFCORE_PIPELINE = False
            utils.NFCORE_PIPELINE_GLOBAL = False
            self.push_screen("basic_details")
        elif event.button.id == "continue":
            self.push_screen("final_details")
        elif event.button.id == "github_repo":
            self.push_screen("github_repo")
        elif event.button.id == "close_screen":
            self.push_screen("github_repo_question")
        elif event.button.id == "exit":
            self.push_screen("github_exit")
        if event.button.id == "close_app":
            self.exit(return_code=0)
        if event.button.id == "back":
            self.pop_screen()

    def action_toggle_dark(self) -> None:
        """An action to toggle dark mode."""
        self.dark: bool = not self.dark

    def action_toggle_all(self) -> None:
        """An action to toggle all Switches."""
        switches = self.query(Switch)
        if not switches:
            return  # No Switches widgets found
        # Determine the new state based on the first switch
        new_state = not switches.first().value if switches.first() else True
        for switch in switches:
            switch.value = new_state
        self.refresh()



import logging
import shutil
import tempfile
from pathlib import Path
from unittest import mock

import questionary
import yaml

import nf_core.utils
from nf_core.components.components_utils import NF_CORE_MODULES_NAME, NF_CORE_MODULES_REMOTE
from nf_core.modules.install import ModuleInstall
from nf_core.modules.modules_json import ModulesJson
from nf_core.modules.patch import ModulePatch
from nf_core.modules.update import ModuleUpdate

from ..test_modules import TestModules
from ..utils import (
    GITLAB_BRANCH_TEST_BRANCH,
    GITLAB_BRANCH_TEST_NEW_SHA,
    GITLAB_BRANCH_TEST_OLD_SHA,
    GITLAB_DEFAULT_BRANCH,
    GITLAB_REPO,
    GITLAB_URL,
    OLD_TRIMGALORE_BRANCH,
    OLD_TRIMGALORE_SHA,
    cmp_component,
)


class TestModulesInstall(TestModules):
    def test_install_and_update(self):
        """Installs a module in the pipeline and updates it (no change)"""
        self.mods_install.install("trimgalore")
        update_obj = ModuleUpdate(self.pipeline_dir, show_diff=False)

        # Copy the module files and check that they are unaffected by the update
        tmpdir = Path(tempfile.TemporaryDirectory().name)
        trimgalore_tmpdir = tmpdir / "trimgalore"
        trimgalore_path = Path(self.pipeline_dir, "modules", NF_CORE_MODULES_NAME, "trimgalore")
        shutil.copytree(trimgalore_path, trimgalore_tmpdir)

        assert update_obj.update("trimgalore") is True
        assert cmp_component(trimgalore_tmpdir, trimgalore_path) is True

    def test_install_at_hash_and_update(self):
        """Installs an old version of a module in the pipeline and updates it"""
        assert self.mods_install_old.install("trimgalore")
        update_obj = ModuleUpdate(
            self.pipeline_dir, show_diff=False, update_deps=True, remote_url=GITLAB_URL, branch=OLD_TRIMGALORE_BRANCH
        )

        # Copy the module files and check that they are affected by the update
        tmpdir = Path(tempfile.TemporaryDirectory().name)
        trimgalore_tmpdir = tmpdir / "trimgalore"
        trimgalore_path = Path(self.pipeline_dir, "modules", GITLAB_REPO, "trimgalore")
        shutil.copytree(trimgalore_path, trimgalore_tmpdir)

        assert update_obj.update("trimgalore") is True
        assert cmp_component(trimgalore_tmpdir, trimgalore_path) is False

        # Check that the modules.json is correctly updated
        mod_json_obj = ModulesJson(self.pipeline_dir)
        mod_json = mod_json_obj.get_modules_json()
        # Get the up-to-date git_sha for the module from the ModulesRepo object
        correct_git_sha = update_obj.modules_repo.get_latest_component_version("trimgalore", "modules")
        current_git_sha = mod_json["repos"][GITLAB_URL]["modules"][GITLAB_REPO]["trimgalore"]["git_sha"]
        assert correct_git_sha == current_git_sha

    # Mock questionary answer: do not update module, only show diffs
    @mock.patch.object(questionary.Question, "unsafe_ask", return_value=True)
    def test_install_at_hash_and_update_limit_output(self, mock_prompt):
        """Installs an old version of a module in the pipeline and updates it with limited output reporting"""
        self.caplog.set_level(logging.INFO)
        assert self.mods_install_old.install("trimgalore")

        update_obj = ModuleUpdate(
            self.pipeline_dir,
            show_diff=True,
            update_deps=True,
            remote_url=GITLAB_URL,
            branch=OLD_TRIMGALORE_BRANCH,
            limit_output=True,
        )
        assert update_obj.update("trimgalore")

        # Check changes not shown for non-.nf files
        assert "Changes in 'trimgalore/meta.yml' but not shown" in self.caplog.text
        # Check changes shown for .nf files
        assert "Changes in 'trimgalore/main.nf'" in self.caplog.text
        for line in self.caplog.text.split("\n"):
            if line.startswith("---"):
                assert line.endswith("main.nf")

    def test_install_at_hash_and_update_and_save_diff_to_file(self):
        """Installs an old version of a module in the pipeline and updates it"""
        self.mods_install_old.install("trimgalore")
        patch_path = Path(self.pipeline_dir, "trimgalore.patch")
        update_obj = ModuleUpdate(
            self.pipeline_dir,
            save_diff_fn=patch_path,
            sha=OLD_TRIMGALORE_SHA,
            remote_url=GITLAB_URL,
            branch=OLD_TRIMGALORE_BRANCH,
        )

        # Copy the module files and check that they are affected by the update
        tmpdir = Path(tempfile.TemporaryDirectory().name)
        trimgalore_tmpdir = tmpdir / "trimgalore"
        trimgalore_path = Path(self.pipeline_dir, "modules", GITLAB_REPO, "trimgalore")
        shutil.copytree(trimgalore_path, trimgalore_tmpdir)

        assert update_obj.update("trimgalore") is True
        assert cmp_component(trimgalore_tmpdir, trimgalore_path) is True

        # TODO: Apply the patch to the module

    def test_install_at_hash_and_update_and_save_diff_to_file_limit_output(self):
        """Installs an old version of a module in the pipeline and updates it"""
        # Install old version of trimgalore
        self.mods_install_old.install("trimgalore")
        patch_path = Path(self.pipeline_dir, "trimgalore.patch")
        # Update saving the differences to a patch file and with `limit_output`
        update_obj = ModuleUpdate(
            self.pipeline_dir,
            save_diff_fn=patch_path,
            remote_url=GITLAB_URL,
            branch=OLD_TRIMGALORE_BRANCH,
            limit_output=True,
        )
        assert update_obj.update("trimgalore")

        # Check that the patch file was created
        assert patch_path.exists(), f"Patch file was not created at {patch_path}"

        # Read the contents of the patch file
        with open(patch_path) as fh:
            patch_content = fh.read()
            # Check changes not shown for non-.nf files
            assert "Changes in 'trimgalore/meta.yml' but not shown" in patch_content
            # Check changes only shown for main.nf
            assert "Changes in 'trimgalore/main.nf'" in patch_content
            for line in patch_content:
                if line.startswith("---"):
                    assert line.endswith("main.nf")

    def test_update_all(self):
        """Updates all modules present in the pipeline"""
        update_obj = ModuleUpdate(self.pipeline_dir, update_all=True, show_diff=False)
        # Get the current modules.json
        assert update_obj.update() is True

        # We must reload the modules.json to get the updated version
        mod_json_obj = ModulesJson(self.pipeline_dir)
        mod_json = mod_json_obj.get_modules_json()
        # Loop through all modules and check that they are updated (according to the modules.json file)
        for mod in mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"][NF_CORE_MODULES_NAME]:
            correct_git_sha = list(update_obj.modules_repo.get_component_git_log(mod, "modules", depth=1))[0]["git_sha"]
            current_git_sha = mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"][NF_CORE_MODULES_NAME][mod]["git_sha"]
            assert correct_git_sha == current_git_sha

    def test_update_with_config_fixed_version(self):
        """Try updating when there are entries in the .nf-core.yml"""
        # Install trimgalore at the latest version
        assert self.mods_install_trimgalore.install("trimgalore")

        # Fix the trimgalore version in the .nf-core.yml to an old version
        update_config = {GITLAB_URL: {GITLAB_REPO: {"trimgalore": OLD_TRIMGALORE_SHA}}}
        config_fn, tools_config = nf_core.utils.load_tools_config(self.pipeline_dir)
        setattr(tools_config, "update", update_config)
        assert config_fn is not None and tools_config is not None  # mypy
        with open(Path(self.pipeline_dir, config_fn), "w") as f:
            yaml.dump(tools_config.model_dump(), f)

        # Update all modules in the pipeline
        update_obj = ModuleUpdate(
            self.pipeline_dir, update_all=True, show_diff=False, remote_url=GITLAB_URL, branch=OLD_TRIMGALORE_BRANCH
        )
        assert update_obj.update() is True

        # Check that the git sha for trimgalore is correctly downgraded
        mod_json = ModulesJson(self.pipeline_dir).get_modules_json()
        assert "trimgalore" in mod_json["repos"][GITLAB_URL]["modules"][GITLAB_REPO]
        assert "git_sha" in mod_json["repos"][GITLAB_URL]["modules"][GITLAB_REPO]["trimgalore"]
        assert mod_json["repos"][GITLAB_URL]["modules"][GITLAB_REPO]["trimgalore"]["git_sha"] == OLD_TRIMGALORE_SHA

    def test_update_with_config_dont_update(self):
        """Try updating when module is to be ignored"""
        # Install an old version of trimgalore
        self.mods_install_old.install("trimgalore")

        # Set the trimgalore field to no update in the .nf-core.yml
        update_config = {GITLAB_URL: {GITLAB_REPO: {"trimgalore": False}}}
        config_fn, tools_config = nf_core.utils.load_tools_config(self.pipeline_dir)
        setattr(tools_config, "update", update_config)
        assert config_fn is not None and tools_config is not None  # mypy
        with open(Path(self.pipeline_dir, config_fn), "w") as f:
            yaml.dump(tools_config.model_dump(), f)

        # Update all modules in the pipeline
        update_obj = ModuleUpdate(
            self.pipeline_dir,
            update_all=True,
            show_diff=False,
            sha=OLD_TRIMGALORE_SHA,
            remote_url=GITLAB_URL,
            branch=OLD_TRIMGALORE_BRANCH,
        )
        assert update_obj.update() is True

        # Check that the git sha for trimgalore is correctly downgraded
        mod_json = ModulesJson(self.pipeline_dir).get_modules_json()
        assert "trimgalore" in mod_json["repos"][GITLAB_URL]["modules"][GITLAB_REPO]
        assert "git_sha" in mod_json["repos"][GITLAB_URL]["modules"][GITLAB_REPO]["trimgalore"]
        assert mod_json["repos"][GITLAB_URL]["modules"][GITLAB_REPO]["trimgalore"]["git_sha"] == OLD_TRIMGALORE_SHA

    def test_update_with_config_fix_all(self):
        """Fix the version of all nf-core modules"""
        self.mods_install_trimgalore.install("trimgalore")

        # Fix the version of all nf-core modules in the .nf-core.yml to an old version
        update_config = {GITLAB_URL: OLD_TRIMGALORE_SHA}
        config_fn, tools_config = nf_core.utils.load_tools_config(self.pipeline_dir)
        setattr(tools_config, "update", update_config)
        assert config_fn is not None and tools_config is not None  # mypy
        with open(Path(self.pipeline_dir, config_fn), "w") as f:
            yaml.dump(tools_config.model_dump(), f)

        # Update all modules in the pipeline
        update_obj = ModuleUpdate(
            self.pipeline_dir, update_all=True, show_diff=False, remote_url=GITLAB_URL, branch=OLD_TRIMGALORE_BRANCH
        )
        assert update_obj.update() is True

        # Check that the git sha for trimgalore is correctly downgraded
        mod_json = ModulesJson(self.pipeline_dir).get_modules_json()
        assert "git_sha" in mod_json["repos"][GITLAB_URL]["modules"][GITLAB_REPO]["trimgalore"]
        assert mod_json["repos"][GITLAB_URL]["modules"][GITLAB_REPO]["trimgalore"]["git_sha"] == OLD_TRIMGALORE_SHA

    def test_update_with_config_no_updates(self):
        """Don't update any nf-core modules"""
        assert self.mods_install_old.install("trimgalore")
        old_mod_json = ModulesJson(self.pipeline_dir).get_modules_json()

        # Fix the version of all nf-core modules in the .nf-core.yml to an old version
        update_config = {GITLAB_URL: False}
        config_fn, tools_config = nf_core.utils.load_tools_config(self.pipeline_dir)
        setattr(tools_config, "update", update_config)
        assert config_fn is not None and tools_config is not None  # mypy
        with open(Path(self.pipeline_dir, config_fn), "w") as f:
            yaml.dump(tools_config.model_dump(), f)

        # Update all modules in the pipeline
        update_obj = ModuleUpdate(
            self.pipeline_dir,
            update_all=True,
            show_diff=False,
            sha=OLD_TRIMGALORE_SHA,
            remote_url=GITLAB_URL,
            branch=OLD_TRIMGALORE_BRANCH,
        )
        assert update_obj.update() is True

        # Check that the git sha for trimgalore is correctly downgraded and none of the modules has changed
        mod_json = ModulesJson(self.pipeline_dir).get_modules_json()
        for module in mod_json["repos"][GITLAB_URL]["modules"][GITLAB_REPO]:
            assert "git_sha" in mod_json["repos"][GITLAB_URL]["modules"][GITLAB_REPO][module]
            assert (
                mod_json["repos"][GITLAB_URL]["modules"][GITLAB_REPO][module]["git_sha"]
                == old_mod_json["repos"][GITLAB_URL]["modules"][GITLAB_REPO][module]["git_sha"]
            )

    def test_update_different_branch_single_module(self):
        """Try updating a module in a specific branch"""
        install_obj = ModuleInstall(
            self.pipeline_dir,
            prompt=False,
            force=False,
            remote_url=GITLAB_URL,
            branch=GITLAB_BRANCH_TEST_BRANCH,
            sha=GITLAB_BRANCH_TEST_OLD_SHA,
        )
        assert install_obj.install("fastp")

        update_obj = ModuleUpdate(
            self.pipeline_dir,
            update_deps=True,
            remote_url=GITLAB_URL,
            branch=GITLAB_BRANCH_TEST_BRANCH,
            show_diff=False,
        )
        update_obj.update("fastp")

        # Verify that the branch entry was updated correctly
        modules_json = ModulesJson(self.pipeline_dir)
        assert (
            modules_json.get_component_branch(self.component_type, "fastp", GITLAB_URL, GITLAB_REPO)
            == GITLAB_BRANCH_TEST_BRANCH
        )
        assert modules_json.get_module_version("fastp", GITLAB_URL, GITLAB_REPO) == GITLAB_BRANCH_TEST_NEW_SHA

    def test_update_different_branch_mixed_modules_main(self):
        """Try updating all modules where MultiQC is installed from main branch"""
        # Install fastp
        assert self.mods_install_gitlab_old.install("fastp")

        # Install MultiQC from gitlab default branch
        assert self.mods_install_gitlab.install("multiqc")

        # Try updating
        update_obj = ModuleUpdate(self.pipeline_dir, update_all=True, show_diff=False)
        assert update_obj.update() is True

        modules_json = ModulesJson(self.pipeline_dir)
        # Verify that the branch entry was updated correctly
        assert (
            modules_json.get_component_branch(self.component_type, "fastp", GITLAB_URL, GITLAB_REPO)
            == GITLAB_BRANCH_TEST_BRANCH
        )
        assert modules_json.get_module_version("fastp", GITLAB_URL, GITLAB_REPO) == GITLAB_BRANCH_TEST_NEW_SHA
        # MultiQC is present in both branches but should've been updated using the 'main' branch
        assert (
            modules_json.get_component_branch(self.component_type, "multiqc", GITLAB_URL, GITLAB_REPO)
            == GITLAB_DEFAULT_BRANCH
        )

    def test_update_different_branch_mix_modules_branch_test(self):
        """Try updating all modules where MultiQC is installed from branch-test branch"""
        # Install multiqc from the branch-test branch
        assert self.mods_install_gitlab_old.install(
            "multiqc"
        )  # Force as the same module is installed from github nf-core modules repo
        modules_json = ModulesJson(self.pipeline_dir)
        update_obj = ModuleUpdate(
            self.pipeline_dir,
            update_all=True,
            show_diff=False,
            remote_url=GITLAB_URL,
            branch=GITLAB_BRANCH_TEST_BRANCH,
            sha=GITLAB_BRANCH_TEST_NEW_SHA,
        )
        assert update_obj.update()

        assert (
            modules_json.get_component_branch(self.component_type, "multiqc", GITLAB_URL, GITLAB_REPO)
            == GITLAB_BRANCH_TEST_BRANCH
        )
        assert modules_json.get_module_version("multiqc", GITLAB_URL, GITLAB_REPO) == GITLAB_BRANCH_TEST_NEW_SHA

    # Mock questionary answer: do not update module, only show diffs
    @mock.patch.object(questionary.Question, "unsafe_ask", return_value=False)
    def test_update_only_show_differences(self, mock_prompt):
        """Try updating all modules showing differences.
        Only show diffs, don't actually save any updated files.
        Check that the sha in modules.json is not changed."""

        # Update modules to a fixed old SHA
        update_old = ModuleUpdate(
            self.pipeline_dir, update_all=True, show_diff=False, sha="5e34754d42cd2d5d248ca8673c0a53cdf5624905"
        )
        update_old.update()

        tmpdir = Path(tempfile.TemporaryDirectory().name)
        shutil.copytree(Path(self.pipeline_dir, "modules", NF_CORE_MODULES_NAME), tmpdir)

        update_obj = ModuleUpdate(self.pipeline_dir, update_all=True, show_diff=True)
        assert ModuleUpdate(self.pipeline_dir, update_all=True, show_diff=True).update()

        mod_json = ModulesJson(self.pipeline_dir).get_modules_json()
        # Loop through all modules and check that they are NOT updated (according to the modules.json file)
        # A module that can be updated but shouldn't is fastqc
        # Module multiqc is already up to date so don't check
        mod = "fastqc"
        non_updated_git_sha = list(update_obj.modules_repo.get_component_git_log(mod, "modules", depth=1))[0]["git_sha"]
        current_git_sha = mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"][NF_CORE_MODULES_NAME][mod]["git_sha"]
        assert non_updated_git_sha != current_git_sha
        assert cmp_component(Path(tmpdir, mod), Path(self.pipeline_dir, "modules", NF_CORE_MODULES_NAME, mod)) is True

    # Mock questionary answer: do not update module, only show diffs
    @mock.patch.object(questionary.Question, "unsafe_ask", return_value=False)
    def test_update_only_show_differences_when_patch(self, mock_prompt):
        """Try updating all modules showing differences when there's a patched module.
        Don't update some of them.
        Check that the sha in modules.json is not changed."""
        modules_json = ModulesJson(self.pipeline_dir)
        update_obj = ModuleUpdate(self.pipeline_dir, update_all=True, show_diff=True)

        # Update modules to a fixed old SHA
        update_old = ModuleUpdate(
            self.pipeline_dir, update_all=True, show_diff=False, sha="5e34754d42cd2d5d248ca8673c0a53cdf5624905"
        )
        assert update_old.update()

        # Modify fastqc module, it will have a patch which will be applied during update
        # We modify fastqc because it's one of the modules that can be updated and there's another one before it (custom/dumpsoftwareversions)
        module_path = Path(self.pipeline_dir, "modules", "nf-core", "fastqc")
        main_path = Path(module_path, "main.nf")
        with open(main_path) as fh:
            lines = fh.readlines()
        for line_index in range(len(lines)):
            if lines[line_index] == "    label 'process_medium'\n":
                lines[line_index] = "    label 'process_low'\n"
                break
        with open(main_path, "w") as fh:
            fh.writelines(lines)
        # Create a patch file
        patch_obj = ModulePatch(self.pipeline_dir)
        patch_obj.patch("fastqc")
        # Check that a patch file with the correct name has been created
        assert "fastqc.diff" in [f.name for f in module_path.glob("*.diff")]

        # Update all modules
        assert update_obj.update() is True

        mod_json = modules_json.get_modules_json()
        # Loop through all modules and check that they are NOT updated (according to the modules.json file)
        # A module that can be updated but shouldn't is fastqc
        # Module multiqc is already up to date so don't check
        mod = "fastqc"
        correct_git_sha = list(update_obj.modules_repo.get_component_git_log(mod, "modules", depth=1))[0]["git_sha"]
        current_git_sha = mod_json["repos"][NF_CORE_MODULES_REMOTE]["modules"][NF_CORE_MODULES_NAME][mod]["git_sha"]
        assert correct_git_sha != current_git_sha

    def test_update_module_with_extra_config_file(self):
        """Try updating a module with a config file"""
        # Install the module
        assert self.mods_install.install("trimgalore")
        # Add a nextflow_test.config file to the module
        trimgalore_path = Path(self.pipeline_dir, "modules", "nf-core", "trimgalore")
        Path(trimgalore_path, "nextflow_test.config").touch()
        with open(Path(trimgalore_path, "nextflow_test.config"), "w") as fh:
            fh.write("params.my_param = 'my_value'\n")
        # Update the module
        update_obj = ModuleUpdate(self.pipeline_dir, show_diff=False)
        assert update_obj.update("trimgalore")
        # Check that the nextflow_test.config file is still there
        assert Path(trimgalore_path, "nextflow_test.config").exists()
        with open(Path(trimgalore_path, "nextflow_test.config")) as fh:
            assert "params.my_param = 'my_value'" in fh.read()



import re
from pathlib import Path


def readme(self):
    """Repository ``README.md`` tests

    The ``README.md`` files for a project are very important and must meet some requirements:

    * Nextflow badge

      * If no Nextflow badge is found, a warning is given
      * If a badge is found but the version doesn't match the minimum version in the config file, the test fails
      * Example badge code:

        .. code-block:: md

           [![Nextflow](https://img.shields.io/badge/nextflow-%E2%89%A50.27.6-brightgreen.svg)](https://www.nextflow.io/)

    .. note:: This badge are a markdown image ``![alt-text](<image URL>)`` *inside* a markdown link ``[markdown image](<link URL>)``, so a bit fiddly to write.

    * Zenodo release

        * If pipeline is released but still contains a 'zenodo.XXXXXXX' tag, the test fails

    To disable this test, add the following to the pipeline's ``.nf-core.yml`` file:

    .. code-block:: yaml
        lint:
            readme: False

    To disable subsets of these tests, add the following to the pipeline's ``.nf-core.yml`` file:

    .. code-block:: yaml

            lint:
                readme:
                    - nextflow_badge
                    - zenodo_release

    """
    passed = []
    warned = []
    failed = []

    # Remove field that should be ignored according to the linting config
    ignore_configs = self.lint_config.get("readme", []) if self.lint_config is not None else []

    with open(Path(self.wf_path, "README.md")) as fh:
        content = fh.read()

    if "nextflow_badge" not in ignore_configs:
        # Check that there is a readme badge showing the minimum required version of Nextflow
        # [![Nextflow](https://img.shields.io/badge/nextflow%20DSL2-%E2%89%A524.04.2-23aa62.svg)](https://www.nextflow.io/)
        # and that it has the correct version
        nf_badge_re = r"\[!\[Nextflow\]\(https://img\.shields\.io/badge/nextflow%20DSL2-!?(?:%E2%89%A5|%3E%3D)([\d\.]+)-23aa62\.svg\)\]\(https://www\.nextflow\.io/\)"
        match = re.search(nf_badge_re, content)
        if match:
            nf_badge_version = match.group(1).strip("'\"")
            try:
                if nf_badge_version != self.minNextflowVersion:
                    raise AssertionError()
            except (AssertionError, KeyError):
                failed.append(
                    f"README Nextflow minimum version badge does not match config. Badge: `{nf_badge_version}`, "
                    f"Config: `{self.minNextflowVersion}`"
                )
            else:
                passed.append(
                    f"README Nextflow minimum version badge matched config. Badge: `{nf_badge_version}`, "
                    f"Config: `{self.minNextflowVersion}`"
                )
        else:
            warned.append("README did not have a Nextflow minimum version badge.")

    if "zenodo_doi" not in ignore_configs:
        # Check that zenodo.XXXXXXX has been replaced with the zendo.DOI
        zenodo_re = r"/zenodo\.X+"
        match = re.search(zenodo_re, content)
        if match:
            warned.append(
                "README contains the placeholder `zenodo.XXXXXXX`. "
                "This should be replaced with the zenodo doi (after the first release)."
            )
        else:
            passed.append("README Zenodo placeholder was replaced with DOI.")

    return {"passed": passed, "warned": warned, "failed": failed}



"""Tests covering the subworkflows commands"""

import json
import unittest
from pathlib import Path

import pytest

import nf_core.modules
import nf_core.modules.install
import nf_core.pipelines.create.create
import nf_core.subworkflows

from .utils import (
    GITLAB_SUBWORKFLOWS_BRANCH,
    GITLAB_SUBWORKFLOWS_ORG_PATH_BRANCH,
    GITLAB_URL,
    OLD_SUBWORKFLOWS_SHA,
    create_tmp_pipeline,
)


def create_modules_repo_dummy(tmp_dir):
    """Create a dummy copy of the nf-core/modules repo"""

    root_dir = Path(tmp_dir, "modules")
    Path(root_dir, "modules").mkdir(parents=True, exist_ok=True)
    Path(root_dir, "subworkflows", "nf-core").mkdir(parents=True, exist_ok=True)
    Path(root_dir, "tests", "config").mkdir(parents=True, exist_ok=True)
    with open(Path(root_dir, ".nf-core.yml"), "w") as fh:
        fh.writelines(["repository_type: modules", "\n", "org_path: nf-core", "\n"])
    # TODO Add a mock here
    subworkflow_create = nf_core.subworkflows.SubworkflowCreate(root_dir, "test_subworkflow", "@author", True)
    subworkflow_create.create()

    # Add dummy content to main.nf.test.snap
    test_snap_path = Path(
        root_dir,
        "subworkflows",
        "nf-core",
        "test_subworkflow",
        "tests",
        "main.nf.test.snap",
    )
    test_snap_path.parent.mkdir(parents=True, exist_ok=True)
    with open(test_snap_path, "w") as fh:
        json.dump(
            {
                "my test": {
                    "content": [
                        {
                            "0": [],
                            "versions": {},
                        }
                    ]
                }
            },
            fh,
            indent=4,
        )

    return root_dir


class TestSubworkflows(unittest.TestCase):
    """Class for subworkflows tests"""

    def setUp(self):
        """Create a new PipelineStructure and Launch objects"""
        self.component_type = "subworkflows"

        # Set up the pipeline structure
        self.tmp_dir, self.template_dir, self.pipeline_name, self.pipeline_dir = create_tmp_pipeline()

        # Set up the nf-core/modules repo dummy
        self.nfcore_modules = create_modules_repo_dummy(self.tmp_dir)

        # Set up install objects
        self.subworkflow_install = nf_core.subworkflows.SubworkflowInstall(self.pipeline_dir, prompt=False, force=False)
        self.subworkflow_install_gitlab = nf_core.subworkflows.SubworkflowInstall(
            self.pipeline_dir,
            prompt=False,
            force=False,
            remote_url=GITLAB_URL,
            branch=GITLAB_SUBWORKFLOWS_BRANCH,
        )
        self.subworkflow_install_gitlab_same_org_path = nf_core.subworkflows.SubworkflowInstall(
            self.pipeline_dir,
            prompt=False,
            force=False,
            remote_url=GITLAB_URL,
            branch=GITLAB_SUBWORKFLOWS_ORG_PATH_BRANCH,
        )
        self.subworkflow_install_old = nf_core.subworkflows.SubworkflowInstall(
            self.pipeline_dir,
            prompt=False,
            force=False,
            sha=OLD_SUBWORKFLOWS_SHA,
        )
        self.subworkflow_install_module_change = nf_core.subworkflows.SubworkflowInstall(
            self.pipeline_dir,
            prompt=False,
            force=False,
            sha="8c343b3c8a0925949783dc547666007c245c235b",
        )
        self.mods_install = nf_core.modules.install.ModuleInstall(self.pipeline_dir, prompt=False, force=True)

        # Set up remove objects
        self.subworkflow_remove = nf_core.subworkflows.SubworkflowRemove(self.pipeline_dir)

    @pytest.fixture(autouse=True)
    def _use_caplog(self, caplog):
        self.caplog = caplog



import logging
import re
from pathlib import Path
from typing import Dict, List

from nf_core.pipelines.lint_utils import ignore_file

log = logging.getLogger(__name__)


class LintConfig:
    def __init__(self, wf_path: str, lint_config: Dict[str, List[str]]):
        self.wf_path = wf_path
        self.lint_config = lint_config

    def lint_file(self, lint_name: str, file_path: Path) -> Dict[str, List[str]]:
        """Lint a file and add the result to the passed or failed list."""

        passed: List[str] = []
        failed: List[str] = []
        ignored: List[str] = []
        ignore_configs: List[str] = []

        fn = Path(self.wf_path, file_path)

        passed, failed, ignored, ignore_configs = ignore_file(lint_name, file_path, Path(self.wf_path))

        error_message = f"`{file_path}` not found"
        # check for partial match in failed or ignored
        if not any(f.startswith(error_message) for f in (failed + ignored)):
            try:
                with open(fn) as fh:
                    config = fh.read()
            except Exception as e:
                return {"failed": [f"Could not parse file: {fn}, {e}"]}

            # find sections with a withName: prefix
            sections = re.findall(r"withName:\s*['\"]?(\w+)['\"]?", config)
            log.debug(f"found sections: {sections}")

            # find all .nf files in the workflow directory
            nf_files = list(Path(self.wf_path).rglob("*.nf"))
            log.debug(f"found nf_files: {[str(f) for f in nf_files]}")

            # check if withName sections are present in config, but not in workflow files
            for section in sections:
                if section not in ignore_configs or section.lower() not in ignore_configs:
                    if not any(section in nf_file.read_text() for nf_file in nf_files):
                        failed.append(
                            f"`{file_path}` contains `withName:{section}`, but the corresponding process is not present in any of the Nextflow scripts."
                        )
                    else:
                        passed.append(f"`{section}` found in `{file_path}` and Nextflow scripts.")
                else:
                    ignored.append(f"``{section}` is ignored.")

        return {"passed": passed, "failed": failed, "ignored": ignored}


def modules_config(self) -> Dict[str, List[str]]:
    """Make sure the conf/modules.config file follows the nf-core template, especially removed sections.

    .. note:: You can choose to ignore this lint tests by editing the file called
        ``.nf-core.yml`` in the root of your pipeline and setting the test to false:

        .. code-block:: yaml

            lint:
                modules_config: False

        To disable this test only for specific modules, you can specify a list of module names.

        .. code-block:: yaml

            lint:
                modules_config:
                    - fastqc

    """

    result = LintConfig(self.wf_path, self.lint_config).lint_file("modules_config", Path("conf", "modules.config"))

    return result


def base_config(self) -> Dict[str, List[str]]:
    """Make sure the conf/base.config file follows the nf-core template, especially removed sections.

    .. note:: You can choose to ignore this lint tests by editing the file called
        ``.nf-core.yml`` in the root of your pipeline and setting the test to false:

        .. code-block:: yaml

            lint:
                base_config: False

    """

    result = LintConfig(self.wf_path, self.lint_config).lint_file("base_config", Path("conf", "base.config"))

    return result



import logging
from textwrap import dedent

from textual.app import ComposeResult
from textual.containers import Center
from textual.screen import Screen
from textual.widgets import Button, Footer, Header, Markdown

log = logging.getLogger(__name__)

github_text_markdown = """
After creating the pipeline template locally, we can create a GitHub repository and push the code to it.

Do you want to create a GitHub repository?
"""


class GithubRepoQuestion(Screen):
    """Ask if the user wants to create a GitHub repository."""

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        yield Markdown(
            dedent(
                """
                # Create GitHub repository
                """
            )
        )
        yield Markdown(dedent(github_text_markdown))
        yield Center(
            Button("Create GitHub repo", id="github_repo", variant="success"),
            Button("Finish without creating a repo", id="exit", variant="primary"),
            classes="cta",
        )
